/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin;

import java.util.Optional;

import org.aopalliance.intercept.MethodInvocation;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;
import org.springframework.data.mapping.PersistentProperty;
import org.springframework.data.mapping.context.PersistentEntities;
import org.springframework.data.repository.support.Repositories;
import org.springframework.data.rest.core.Path;
import org.springframework.data.rest.core.config.RepositoryRestConfiguration;
import org.springframework.data.rest.core.mapping.PropertyAwareResourceMapping;
import org.springframework.data.rest.core.mapping.RepositoryResourceMappings;
import org.springframework.data.rest.core.mapping.ResourceDescription;
import org.springframework.data.rest.core.mapping.ResourceMapping;
import org.springframework.data.rest.core.mapping.ResourceMetadata;
import org.springframework.data.rest.core.mapping.SearchResourceMappings;
import org.springframework.data.rest.core.mapping.SupportedHttpMethods;
import org.springframework.hateoas.LinkRelation;
import org.springframework.security.authorization.AuthorizationDecision;
import org.springframework.security.authorization.method.SecuredAuthorizationManager;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.util.SimpleMethodInvocation;

import de.ozgcloud.common.errorhandling.TechnicalException;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Configuration
class OzgCloudRestLinksConfiguration {
	@Primary
	@Bean
	RepositoryResourceMappings ozgCloudResourceMappings(Repositories repositories, PersistentEntities persistentEntities,
			RepositoryRestConfiguration repositoryRestConfiguration) {
		return new OzgCloudDelegatingRepositoryResourceMappings(repositories, persistentEntities, repositoryRestConfiguration);
	}
}

@Log4j2
class OzgCloudDelegatingRepositoryResourceMappings extends RepositoryResourceMappings {

	private final Repositories repositories;

	public OzgCloudDelegatingRepositoryResourceMappings(Repositories repositories, PersistentEntities entities,
			RepositoryRestConfiguration configuration) {
		super(repositories, entities, configuration);

		this.repositories = repositories;
	}

	@Override
	public ResourceMetadata getMetadataFor(Class<?> type) {
		return new OzgCloudDelegatingResourceMetadata(super.getMetadataFor(type), repositories);
	}

	@RequiredArgsConstructor
	static class OzgCloudDelegatingResourceMetadata implements ResourceMetadata {

		private final ResourceMetadata metadata;
		private final Repositories repositories;
		private final SecuredAuthorizationManager authManager = new SecuredAuthorizationManager();

		@Override
		public LinkRelation getItemResourceRel() {
			return metadata.getItemResourceRel();
		}

		@Override
		public ResourceDescription getItemResourceDescription() {
			return metadata.getItemResourceDescription();
		}

		@Override
		public Optional<Class<?>> getExcerptProjection() {
			return metadata.getExcerptProjection();
		}

		@Override
		public boolean isExported() {
			return isAccessPermitted(metadata.getDomainType()) && metadata.isExported();
		}

		boolean isAccessPermitted(Class<?> type) {
			var repository = repositories.getRepositoryFor(type);
			return repository.map(repo -> authManager.check(() -> SecurityContextHolder.getContext().getAuthentication(), getFindAllInvocation(repo)))
					.map(AuthorizationDecision::isGranted)
					.orElse(true);
		}

		MethodInvocation getFindAllInvocation(Object repository) {
			try {
				var method = repository.getClass().getMethod("findAll");
				return new SimpleMethodInvocation(repository, method);
			} catch (NoSuchMethodException | SecurityException e) {
				LOG.error(e.getMessage(), e);
				throw new TechnicalException("Error on extracting findAll Repository Method", e);
			}
		}

		@Override
		public LinkRelation getRel() {
			return metadata.getRel();
		}

		@Override
		public Path getPath() {
			return metadata.getPath();
		}

		@Override
		public boolean isPagingResource() {
			return metadata.isPagingResource();
		}

		@Override
		public ResourceDescription getDescription() {
			return metadata.getDescription();
		}

		@Override
		public Class<?> getDomainType() {
			return metadata.getDomainType();
		}

		@Override
		public boolean isExported(PersistentProperty<?> property) {
			return metadata.isExported(property);
		}

		@Override
		public PropertyAwareResourceMapping getProperty(String mappedPath) {
			return metadata.getProperty(mappedPath);
		}

		@Override
		public ResourceMapping getMappingFor(PersistentProperty<?> property) {
			return metadata.getMappingFor(property);
		}

		@Override
		public SearchResourceMappings getSearchResourceMappings() {
			return metadata.getSearchResourceMappings();
		}

		@Override
		public SupportedHttpMethods getSupportedHttpMethods() {
			return metadata.getSupportedHttpMethods();
		}
	}
}
