/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.common.user;

import java.util.Collection;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import org.springframework.security.authentication.AuthenticationTrustResolver;
import org.springframework.security.authentication.AuthenticationTrustResolverImpl;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.context.SecurityContextHolder;

import lombok.AccessLevel;
import lombok.NoArgsConstructor;

@NoArgsConstructor(access = AccessLevel.PRIVATE)
class CurrentUserHelper {
	static final String ROLE_PREFIX = "ROLE_";
	private static final Predicate<String> IS_ROLE_PREFIX_MISSING = role -> !role.startsWith(ROLE_PREFIX);
	private static final AuthenticationTrustResolver TRUST_RESOLVER = new AuthenticationTrustResolverImpl();
	private static final Predicate<Authentication> IS_TRUSTED = auth -> !TRUST_RESOLVER.isAnonymous(auth);

	public static boolean hasRole(String role) {
		return hasAnyRole(Set.of(role));
	}

	public static boolean hasAnyRole(Set<String> roles) {
		var auth = getAuthentication();

		if ((Objects.isNull(auth)) || (Objects.isNull(auth.getPrincipal()))) {
			return false;
		}

		return containsAnyRole(auth.getAuthorities(), roles);
	}

	static boolean containsAnyRole(Collection<? extends GrantedAuthority> authorities, Set<String> roles) {
		if (Objects.isNull(authorities)) {
			return false;
		}
		var rolesWithPrefix = roles.stream().map(CurrentUserHelper::addRolePrefixIfMissing).collect(Collectors.toSet());
		return authorities.stream().anyMatch(a -> rolesWithPrefix.contains(a.getAuthority()));
	}

	static String addRolePrefixIfMissing(String roleToCheck) {
		return Optional.ofNullable(roleToCheck)
				.filter(IS_ROLE_PREFIX_MISSING)
				.map(role -> ROLE_PREFIX + role)
				.orElse(roleToCheck);
	}

	static Authentication getAuthentication() {
		return findAuthentication().orElseThrow(() -> new IllegalStateException("No authenticated User found"));
	}

	private static Optional<Authentication> findAuthentication() {
		return Optional.ofNullable(SecurityContextHolder.getContext().getAuthentication()).filter(IS_TRUSTED);
	}
}
