/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.organisationseinheit;

import java.util.List;
import java.util.Objects;

import org.springframework.data.rest.webmvc.ResourceNotFoundException;
import org.springframework.stereotype.Service;

import de.ozgcloud.common.errorhandling.TechnicalException;
import lombok.RequiredArgsConstructor;

@Service
@RequiredArgsConstructor
class OrganisationsEinheitService {

	private final OrganisationsEinheitRepository repository;

	public OrganisationsEinheit saveOrganisationsEinheit(OrganisationsEinheit organisationsEinheit) {
		return repository.save(organisationsEinheit);
	}

	public List<OrganisationsEinheit> getOrganisationsEinheiten() {
		var organisationsEinheiten = repository.findAllNotDeleted();
		return organisationsEinheiten.stream()
				.filter(organisationsEinheit -> Objects.isNull(organisationsEinheit.getParentId()))
				.map(rootOrganisationsEinheit -> addChildren(rootOrganisationsEinheit, organisationsEinheiten))
				.toList();
	}

	OrganisationsEinheit addChildren(OrganisationsEinheit rootOrganisationsEinheit, List<OrganisationsEinheit> allOrganisationsEinheiten) {
		return rootOrganisationsEinheit.toBuilder()
				.children(allOrganisationsEinheiten.stream()
						.filter(organisationsEinheit -> rootOrganisationsEinheit.getId().equals(organisationsEinheit.getParentId())).toList())
				.build();
	}

	public OrganisationsEinheit getOrganisationsEinheitById(String id) {
		return repository.findById(id).orElseThrow(() -> new ResourceNotFoundException("Organisationseinheit with id " + id + " not found"));
	}

	public List<OrganisationsEinheit> getChildren(String parentId) {
		return repository.findChildren(parentId);
	}

	public OrganisationsEinheit editOrganisationsEinheit(String id, OrganisationsEinheitPatch organisationsEinheitPatch) {
		return repository.findById(id)
				.map(organisationsEinheit -> patchOrganisationsEinheit(organisationsEinheit, organisationsEinheitPatch))
				.map(repository::save)
				.orElseThrow(() -> new TechnicalException("Organisationseinheit with id " + id + " not found"));
	}

	OrganisationsEinheit patchOrganisationsEinheit(OrganisationsEinheit organisationsEinheit, OrganisationsEinheitPatch patch) {
		return organisationsEinheit.toBuilder()
				.settings(
						organisationsEinheit.getSettings().toBuilder()
								.signatur(patch.getSignatur())
								.build())
				.build();
	}
}
