/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.organisationseinheit;

import java.util.List;
import java.util.Optional;

import org.springframework.stereotype.Service;

import de.ozgcloud.admin.keycloak.AddGroupData;
import de.ozgcloud.admin.keycloak.Group;
import de.ozgcloud.admin.keycloak.KeycloakRemoteService;
import de.ozgcloud.admin.keycloak.ResourceCreationException;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Service
@RequiredArgsConstructor
@Log4j2
class SyncService {

	private final OrganisationsEinheitRepository repository;
	private final KeycloakRemoteService keycloakRemoteService;
	private final OrganisationsEinheitRemoteService organisationsRemoteService;
	private final OrganisationsEinheitMapper organisationsEinheitMapper;

	public void syncOrganisationsEinheitenFromKeycloak(long syncTimestamp) {
		keycloakRemoteService.getGroupsWithOrganisationsEinheitId().forEach(group -> syncGroupsWithSubGroups(group, null, syncTimestamp));
		repository.setUnsyncedAsDeleted(syncTimestamp);
	}

	void syncGroupsWithSubGroups(Group group, OrganisationsEinheit parent, long syncTimestamp) {
		var synced = syncGroup(group, parent, syncTimestamp);
		var saved = saveSyncedOrganisationsEinheit(synced);
		group.getSubGroups().forEach(subGroup -> syncGroupsWithSubGroups(subGroup, saved, syncTimestamp));
	}

	OrganisationsEinheit syncGroup(Group group, OrganisationsEinheit parent, long syncTimestamp) {
		var pvogOrganisationsEinheiten = organisationsRemoteService.getByOrganisationsEinheitId(group.getOrganisationsEinheitId());
		var syncedName = syncName(pvogOrganisationsEinheiten, group);
		var syncResult = evaluateSyncResult(pvogOrganisationsEinheiten, group);
		var zufiId = syncZufiId(pvogOrganisationsEinheiten);
		var organisationsEinheitBuilder = OrganisationsEinheit.builder()
				.keycloakId(group.getId())
				.name(syncedName)
				.organisationsEinheitId(group.getOrganisationsEinheitId())
				.syncResult(syncResult)
				.zufiId(zufiId)
				.lastSyncTimestamp(syncTimestamp);
		if (parent != null) {
			organisationsEinheitBuilder.parentId(parent.getId());
		}
		return organisationsEinheitBuilder.build();
	}

	String syncName(List<OrganisationsEinheit> pvogOrganisationsEinheiten, Group group) {
		if (pvogOrganisationsEinheiten.size() != 1) {
			return group.getName();
		}
		return pvogOrganisationsEinheiten.getFirst().getName();
	}

	SyncResult evaluateSyncResult(List<OrganisationsEinheit> pvogOrganisationsEinheiten, Group group) {
		if (pvogOrganisationsEinheiten.isEmpty()) {
			return SyncResult.NOT_FOUND_IN_PVOG;
		}
		if (pvogOrganisationsEinheiten.size() > 1) {
			return SyncResult.ORGANISATIONSEINHEIT_ID_NOT_UNIQUE;
		}
		if (!pvogOrganisationsEinheiten.getFirst().getName().equals(group.getName())) {
			return SyncResult.NAME_MISMATCH;
		}
		return SyncResult.OK;
	}

	public String syncZufiId(List<OrganisationsEinheit> pvogOrganisationsEinheiten) {
		if (pvogOrganisationsEinheiten.size() != 1) {
			return null;
		}
		return pvogOrganisationsEinheiten.getFirst().getId();
	}

	OrganisationsEinheit saveSyncedOrganisationsEinheit(OrganisationsEinheit syncedOrganisationsEinheit) {
		var existingOrganisationsEinheit = repository.findSyncedByKeycloakId(syncedOrganisationsEinheit.getKeycloakId());

		return repository.save(existingOrganisationsEinheit
				.map(organisationsEinheit -> syncedOrganisationsEinheit.toBuilder()
						.id(organisationsEinheit.getId())
						.settings(organisationsEinheit.getSettings()).build())
				.orElse(syncedOrganisationsEinheit));
	}

	public void syncAddedOrganisationsEinheiten(long syncTimestamp) {
		repository.findAllWithoutSyncResult().forEach(
				organisationsEinheit -> syncAddedOrganisationsEinheit(organisationsEinheit, syncTimestamp));
	}

	void syncAddedOrganisationsEinheit(OrganisationsEinheit organisationsEinheit, long syncTimestamp) {
		addAsGroupInKeycloak(organisationsEinheit).ifPresent(
				addedGroupId -> updateAfterSuccessfulGroupCreation(organisationsEinheit, addedGroupId, syncTimestamp));
	}

	Optional<String> addAsGroupInKeycloak(OrganisationsEinheit organisationsEinheit) {
		if (organisationsEinheit.getParentId() != null) {
			throw new OrganisationsEinheitSynchronizationException(
					"Organisationseinheit %s has parent".formatted(organisationsEinheit.getOrganisationsEinheitId()));
		}
		var addGroupData = organisationsEinheitMapper.toAddGroupData(organisationsEinheit);
		return addGroupInKeycloak(addGroupData);
	}

	void updateAfterSuccessfulGroupCreation(OrganisationsEinheit organisationsEinheit, String keycloakId, long syncTimestamp) {
		var updatedOrganisationsEinheit = organisationsEinheit.toBuilder()
				.keycloakId(keycloakId)
				.syncResult(SyncResult.OK)
				.lastSyncTimestamp(syncTimestamp)
				.build();
		repository.save(updatedOrganisationsEinheit);
	}

	Optional<String> addGroupInKeycloak(AddGroupData addGroupData) {
		try {
			return Optional.of(keycloakRemoteService.addGroup(addGroupData));
		} catch (ResourceCreationException e) {
			throw new OrganisationsEinheitSynchronizationException("Error creating group %s in Keycloak".formatted(addGroupData), e);
		}
	}
}
