/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.security;

import static java.util.stream.Collectors.*;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.http.HttpMethod;
import org.springframework.security.config.Customizer;
import org.springframework.security.config.annotation.method.configuration.EnableMethodSecurity;
import org.springframework.security.config.annotation.web.builders.HttpSecurity;
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity;
import org.springframework.security.config.http.SessionCreationPolicy;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.oauth2.core.oidc.StandardClaimNames;
import org.springframework.security.oauth2.jwt.Jwt;
import org.springframework.security.oauth2.server.resource.authentication.JwtAuthenticationConverter;
import org.springframework.security.web.SecurityFilterChain;

import de.ozgcloud.admin.common.user.UserRole;
import de.ozgcloud.admin.environment.OAuth2Properties;
import lombok.RequiredArgsConstructor;

@Configuration
@EnableMethodSecurity(securedEnabled = true)
@EnableWebSecurity
@RequiredArgsConstructor
public class SecurityConfiguration {

	static final String RESOURCE_ACCESS_KEY = "resource_access";
	static final String SIMPLE_GRANT_AUTHORITY_PREFIX = "ROLE_";
	static final String ROLES_KEY = "roles";

	private final AdminAuthenticationEntryPoint authenticationEntryPoint;
	private final OAuth2Properties oAuth2Properties;

	@Bean
	SecurityFilterChain filterChain(HttpSecurity http) throws Exception {

		http.oauth2ResourceServer(oauth2 -> oauth2.jwt(Customizer.withDefaults()));

		http.sessionManagement(sm -> sm.sessionCreationPolicy(SessionCreationPolicy.STATELESS));

		http.exceptionHandling(eh -> eh.authenticationEntryPoint(authenticationEntryPoint));

		http.authorizeHttpRequests(requests -> requests
				.requestMatchers(HttpMethod.GET, "/api/environment").permitAll()
				.requestMatchers("/api/configuration").hasAnyRole(UserRole.ADMIN_ADMIN, UserRole.DATENBEAUFTRAGUNG)
				.requestMatchers("/api/configuration/**").hasAnyRole(UserRole.ADMIN_ADMIN, UserRole.DATENBEAUFTRAGUNG)
				.requestMatchers("/api").authenticated()
				.requestMatchers("/api/**").authenticated()
				.requestMatchers("/actuator").permitAll()
				.requestMatchers("/actuator/**").permitAll()
				.requestMatchers("/configserver/**").permitAll()
				.anyRequest().denyAll());

		return http.build();
	}

	// TODO OZG-4954 replace with spring defaults
	@Bean
	JwtAuthenticationConverter jwtAuthenticationConverter() {
		var jwtConverter = new JwtAuthenticationConverter();
		jwtConverter.setJwtGrantedAuthoritiesConverter(this::convertJwtToGrantedAuthorities);
		jwtConverter.setPrincipalClaimName(StandardClaimNames.PREFERRED_USERNAME);
		return jwtConverter;
	}

	Set<GrantedAuthority> convertJwtToGrantedAuthorities(Jwt jwt) {
		return getRolesFromJwt(jwt)
				.stream()
				.map(this::mapRoleStringToGrantedAuthority)
				.collect(toSet());
	}

	private GrantedAuthority mapRoleStringToGrantedAuthority(String role) {
		return new SimpleGrantedAuthority(SIMPLE_GRANT_AUTHORITY_PREFIX + role);
	}

	List<String> getRolesFromJwt(Jwt jwt) {
		return Optional.ofNullable(jwt.getClaimAsMap(RESOURCE_ACCESS_KEY))
				.flatMap(resourceAccessMap -> getMap(resourceAccessMap, oAuth2Properties.getResource()))
				.flatMap(adminClientMap -> getList(adminClientMap, ROLES_KEY))
				.orElse(Collections.emptyList());
	}

	@SuppressWarnings("unchecked")
	private Optional<Map<String, Object>> getMap(Map<String, Object> map, String mapKey) {
		return Optional.ofNullable(map.get(mapKey))
				.filter(Map.class::isInstance)
				.map(obj -> (Map<String, Object>) obj);
	}

	@SuppressWarnings("unchecked")
	private Optional<List<String>> getList(Map<String, Object> map, String mapKey) {
		return Optional.ofNullable(map.get(mapKey))
				.filter(List.class::isInstance)
				.map(obj -> (List<String>) obj);
	}

}
