/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.setting;

import static de.ozgcloud.admin.setting.SettingConstants.*;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import jakarta.annotation.PostConstruct;

import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.ListableBeanFactory;
import org.springframework.stereotype.Service;

import static java.util.stream.Collectors.*;

import de.ozgcloud.admin.common.DtoService;
import de.ozgcloud.admin.common.SettingDtoService;
import de.ozgcloud.admin.organisationseinheit.OrganisationsEinheitSettings;
import de.ozgcloud.admin.setting.postfach.PostfachSettingBody;
import lombok.RequiredArgsConstructor;

@Service
@RequiredArgsConstructor
class SettingService {
	private static final String POSTFACH_SETTING_ITEM_NAME = "Postfach";

	private final OrganisationsEinheitSettingsRepository organisationsEinheitSettingsRepository;
	private final SettingRepository repository;

	private final SettingMapper mapper;
	private final ListableBeanFactory beanFactory;

	private Map<String, List<SettingDtoService>> dtoServiceByManager;

	@PostConstruct
	void initDtoServices() {
		this.dtoServiceByManager = beanFactory.getBeansWithAnnotation(DtoService.class).values().stream()
				.map(SettingDtoService.class::cast)
				.flatMap(this::getManagerForService)
				.collect(Collectors.groupingBy(Pair::getKey, mapping(Pair::getValue, toList())));
	}

	private Stream<Pair<String, SettingDtoService>> getManagerForService(SettingDtoService dtoService) {
		var annotation = dtoService.getClass().getDeclaredAnnotation(DtoService.class);
		return Arrays.stream(annotation.value()).map(manager -> Pair.of(manager, dtoService));
	}

	public AlfaSettingDTO getAlfaSettingDTO() {
		var postfachData = getSettingWithPostfachFromDb();
		Map<String, OrganisationsEinheitSettings> settings = getOrganisationsEinheitSettings();

		return AlfaSettingDTO.builder()
				.signatur(postfachData.getSignatur())
				.organisationsEinheitSettings(settings)
				.build();
	}

	public VorgangManagerSettingDTO getVorgangManagerSettingDTO() {
		var postfachData = getSettingWithPostfachFromDb();
		Map<String, OrganisationsEinheitSettings> settings = getOrganisationsEinheitSettings();

		return VorgangManagerSettingDTO.builder()
				.absender(mapper.mapAbsenderToAbsenderVorgangManager(postfachData.getAbsender()))
				.signatur(postfachData.getSignatur())
				.organisationsEinheitSettings(settings)
				.build();
	}

	PostfachSettingBody getSettingWithPostfachFromDb() {
		return repository.findOneByName(POSTFACH_SETTING_ITEM_NAME)
				.map(Setting::getSettingBody)
				.map(PostfachSettingBody.class::cast)
				.orElse(PostfachSettingBody.builder().build());
	}

	Map<String, OrganisationsEinheitSettings> getOrganisationsEinheitSettings() {
		return organisationsEinheitSettingsRepository.getSettings().collect(Collectors.toMap(
				OrganisationsEinheitWithSettings::organisationsEinheitId,
				OrganisationsEinheitWithSettings::settings));
	}

	public ApplicationSettingDTO getAggregationManagerSettingDto() {
		return MapBasedApplicationSettingDto.builder()
				.settings(buildSettingsMap(AGGREGATION_MANAGER))
				.build();
	}

	private Map<String, Object> buildSettingsMap(String managerName) {
		return getDtoService(managerName).stream()
				.collect(Collectors.toMap(SettingDtoService::getSettingKeyName, SettingDtoService::getDataTransferObject));
	}

	private List<SettingDtoService> getDtoService(String managerName) {
		return dtoServiceByManager.getOrDefault(managerName, Collections.emptyList());
	}

}
