package de.ozgcloud.admin;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;

import org.aopalliance.intercept.MethodInvocation;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.Spy;
import org.springframework.data.mongodb.repository.MongoRepository;
import org.springframework.data.repository.support.Repositories;
import org.springframework.data.rest.core.mapping.ResourceMetadata;
import org.springframework.security.authorization.AuthorizationDecision;
import org.springframework.security.authorization.method.SecuredAuthorizationManager;

import de.ozgcloud.admin.OzgCloudDelegatingRepositoryResourceMappings.OzgCloudDelegatingResourceMetadata;
import de.ozgcloud.common.test.ReflectionTestUtils;

class OzgCloudDelegatingResourceMetadataTest {

	@Spy
	@InjectMocks
	private OzgCloudDelegatingResourceMetadata delegatingResourceMetadata;

	@Mock
	private ResourceMetadata metadata;
	@Mock
	private Repositories repositories;
	@Mock
	private SecuredAuthorizationManager authManager;

	@Nested
	class TestIsExported {

		@Test
		void shouldReturnTrue() {
			when(metadata.isExported()).thenReturn(true);
			doReturn(true).when(delegatingResourceMetadata).isAccessPermitted(any());

			var result = delegatingResourceMetadata.isExported();

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseIfNotPermitted() {
			doReturn(false).when(delegatingResourceMetadata).isAccessPermitted(any());

			var result = delegatingResourceMetadata.isExported();

			assertThat(result).isFalse();
		}

		@Test
		void shouldReturnFalseIfNotExported() {
			doReturn(true).when(delegatingResourceMetadata).isAccessPermitted(any());
			when(metadata.isExported()).thenReturn(false);

			var result = delegatingResourceMetadata.isExported();

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestIsAccessPermitted {
		@Mock
		private Object repository;
		@Mock
		private AuthorizationDecision decision;

		private final MethodInvocation invocation = Mockito.mock(MethodInvocation.class);

		@BeforeEach
		void initMock() {
			when(repositories.getRepositoryFor(any())).thenReturn(Optional.of(repository));

			doReturn(invocation).when(delegatingResourceMetadata).getFindAllInvocation(any());
		}

		@BeforeEach
		void mockAuthManager() {
			when(authManager.check(any(), any())).thenReturn(decision);

			ReflectionTestUtils.setField(delegatingResourceMetadata, "authManager", authManager);
		}

		@Test
		void shouldCallDecisionManager() {
			delegatingResourceMetadata.isAccessPermitted(Object.class);

			verify(decision).isGranted();
		}

		@Test
		void shouldReturnTrueIfGranted() {
			when(decision.isGranted()).thenReturn(true);

			var result = delegatingResourceMetadata.isAccessPermitted(any());

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseIfNotGranted() {
			when(decision.isGranted()).thenReturn(false);

			var result = delegatingResourceMetadata.isAccessPermitted(any());

			assertThat(result).isFalse();
		}

		@Test
		void shouldCallAuthManager() {
			delegatingResourceMetadata.isAccessPermitted(Object.class);

			verify(authManager).check(any(), eq(invocation));
		}

		@Test
		void shouldReturnTrueIfNotSecured() {
			when(authManager.check(any(), any())).thenReturn(null);

			var result = delegatingResourceMetadata.isAccessPermitted(any());

			assertThat(result).isTrue();
		}
	}

	@Nested
	class TestGetFindAllInvocation {

		@Mock
		private MongoRepository<String, Object> repository;

		@Test
		void shouldReturnMethod() {
			var method = delegatingResourceMetadata.getFindAllInvocation(repository);

			assertThat(method).isNotNull()
					.extracting(invocation -> invocation.getMethod().getName()).isEqualTo("findAll");
		}
	}

}
