/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.common.errorhandling;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.Collections;

import jakarta.validation.ConstraintViolationException;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.data.rest.webmvc.ResourceNotFoundException;
import org.springframework.http.HttpStatus;
import org.springframework.http.ProblemDetail;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import com.jayway.jsonpath.JsonPath;
import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.common.errorhandling.TechnicalException;
import lombok.SneakyThrows;

class ExceptionControllerTest {

	@InjectMocks
	private ExceptionController exceptionController;

	@Mock
	private ProblemDetailMapper problemDetailMapper;

	private MockMvc mockMvc;

	@BeforeEach
	void setup() {
		mockMvc = MockMvcBuilders
				.standaloneSetup(new TestErrorController())
				.setControllerAdvice(exceptionController).build();
	}

	@DisplayName("Runtime Exception")
	@Nested
	class TestRuntimeException {

		private final static String TEST_ERROR_ENDPOINT = TestErrorController.BASE_PATH + "/runtime";
		private final static int STATUS_VALUE = HttpStatus.INTERNAL_SERVER_ERROR.value();

		@SneakyThrows
		@Test
		void shouldReturnStatus() {
			var result = doPerformGet();

			result.andExpect(status().is(STATUS_VALUE));
		}

		@DisplayName("response body")
		@Nested
		class TestRuntimeExceptionBody {

			@SneakyThrows
			@Test
			void shouldHaveStatus() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.status").value(STATUS_VALUE));
			}

			@SneakyThrows
			@Test
			void shouldHaveTitle() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.title").value("Internal Server Error"));
			}

			@SneakyThrows
			@Test
			void shouldHaveDetail() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.detail").value("error message"));
			}

			@SneakyThrows
			@Test
			void shouldHaveInstance() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.instance").value(TEST_ERROR_ENDPOINT));
			}

			@SneakyThrows
			void shouldHaveType() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.type").value("about:blank"));
			}
		}

		@SneakyThrows
		private ResultActions doPerformGet() {
			return mockMvc.perform(get(TEST_ERROR_ENDPOINT));
		}
	}

	@DisplayName("Access Denied Exception")
	@Nested
	class TestAccessDeniedException {

		private final static String TEST_ERROR_ENDPOINT = TestErrorController.BASE_PATH + "/access-denied";
		private final static int STATUS_VALUE = HttpStatus.FORBIDDEN.value();

		@SneakyThrows
		@Test
		void shouldReturnStatus() {
			var result = doPerformGet();

			result.andExpect(status().is(STATUS_VALUE));
		}

		@DisplayName("response body")
		@Nested
		class TestAccessDeniedExceptionBody {

			@SneakyThrows
			@Test
			void shouldHaveStatus() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.status").value(STATUS_VALUE));
			}

			@SneakyThrows
			@Test
			void shouldHaveTitle() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.title").value("Forbidden"));
			}

			@SneakyThrows
			@Test
			void shouldHaveDetail() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.detail").value("error message"));
			}

			@SneakyThrows
			@Test
			void shouldHaveInstance() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.instance").value(TEST_ERROR_ENDPOINT));
			}

			@SneakyThrows
			void shouldHaveType() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.type").value("about:blank"));
			}
		}

		@SneakyThrows
		private ResultActions doPerformGet() {
			return mockMvc.perform(get(TEST_ERROR_ENDPOINT));
		}
	}

	@DisplayName("ResourceNotFound Exception")
	@Nested
	class TestResourceNotFoundException {

		private final static String TEST_ERROR_ENDPOINT = TestErrorController.BASE_PATH + "/resource-not-found";
		private final static int STATUS_VALUE = HttpStatus.NOT_FOUND.value();

		@SneakyThrows
		@Test
		void shouldReturnStatus() {
			var result = doPerformGet();

			result.andExpect(status().is(STATUS_VALUE));
		}

		@DisplayName("response body")
		@Nested
		class TestResourceNotFoundExceptionBody {

			@SneakyThrows
			@Test
			void shouldHaveStatus() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.status").value(STATUS_VALUE));
			}

			@SneakyThrows
			@Test
			void shouldHaveTitle() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.title").value("Not Found"));
			}

			@SneakyThrows
			@Test
			void shouldHaveDetail() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.detail").value("error message"));
			}

			@SneakyThrows
			@Test
			void shouldHaveInstance() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.instance").value(TEST_ERROR_ENDPOINT));
			}

			@SneakyThrows
			void shouldHaveType() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.type").value("about:blank"));
			}
		}

		@SneakyThrows
		private ResultActions doPerformGet() {
			return mockMvc.perform(get(TEST_ERROR_ENDPOINT));
		}
	}

	@DisplayName("Functional Exception")
	@Nested
	class TestFunctionalException {

		private final static String TEST_ERROR_ENDPOINT = TestErrorController.BASE_PATH + "/functional";
		private final static int STATUS_VALUE = HttpStatus.BAD_REQUEST.value();

		@SneakyThrows
		@Test
		void shouldReturnStatus() {
			var result = doPerformGet();

			result.andExpect(status().is(STATUS_VALUE));
		}

		@DisplayName("response body")
		@Nested
		class TestFunctionalExceptionBody {

			@SneakyThrows
			@Test
			void shouldHaveStatus() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.status").value(STATUS_VALUE));
			}

			@SneakyThrows
			@Test
			void shouldHaveTitle() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.title").value("Bad Request"));
			}

			@DisplayName("detail")
			@Nested
			class TestBodyDetail {

				@SneakyThrows
				@Test
				void shouldContainErrorMessage() {
					var result = doPerformGet();

					assertTrue(getDetailFromResponseContent(result).contains("Functional error: error message"));
				}

				@SneakyThrows
				@Test
				void shouldContainExceptionId() {
					var result = doPerformGet();

					assertTrue(getDetailFromResponseContent(result).contains("(ExceptionId: "));
				}
			}

			@SneakyThrows
			@Test
			void shouldHaveInstance() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.instance").value(TEST_ERROR_ENDPOINT));
			}

			@SneakyThrows
			void shouldHaveType() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.type").value("about:blank"));
			}
		}

		@SneakyThrows
		private ResultActions doPerformGet() {
			return mockMvc.perform(get(TEST_ERROR_ENDPOINT));
		}
	}

	@DisplayName("TechnicalException")
	@Nested
	class TestTechnicalException {

		private final static String TEST_ERROR_ENDPOINT = TestErrorController.BASE_PATH + "/technical";
		private final static int STATUS_VALUE = HttpStatus.INTERNAL_SERVER_ERROR.value();

		@SneakyThrows
		@Test
		void shouldReturnStatus() {
			var result = doPerformGet();

			result.andExpect(status().is(STATUS_VALUE));
		}

		@DisplayName("response body")
		@Nested
		class TestTechnicalExceptionBody {

			@SneakyThrows
			@Test
			void shouldHaveStatus() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.status").value(STATUS_VALUE));
			}

			@SneakyThrows
			@Test
			void shouldHaveTitle() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.title").value("Internal Server Error"));
			}

			@DisplayName("detail")
			@Nested
			class TestBodyDetail {

				@SneakyThrows
				@Test
				void shouldContainErrorMessage() {
					var result = doPerformGet();

					assertTrue(getDetailFromResponseContent(result).contains("error message"));
				}

				@SneakyThrows
				@Test
				void shouldContainExceptionId() {
					var result = doPerformGet();

					assertTrue(getDetailFromResponseContent(result).contains("(ExceptionId: "));
				}
			}

			@SneakyThrows
			@Test
			void shouldHaveInstance() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.instance").value(TEST_ERROR_ENDPOINT));
			}

			@SneakyThrows
			void shouldHaveType() {
				var result = doPerformGet();

				result.andExpect(jsonPath("$.type").value("about:blank"));
			}
		}

		@SneakyThrows
		private ResultActions doPerformGet() {
			return mockMvc.perform(get(TEST_ERROR_ENDPOINT));
		}
	}

	@Nested
	class TestHandleConstraintViolationException {
		private final String exceptionMessage = LoremIpsum.getInstance().getWords(5);

		private final ConstraintViolationException exception = new ConstraintViolationException(exceptionMessage,
				Collections.singleton(ConstraintViolationTestFactory.buildMockedConstraintViolation()));

		@Test
		void shouldBuildConstraintViolationProblemDetail() {
			handleException();

			verify(problemDetailMapper).fromConstraintViolationException(exception);
		}

		@Test
		void shouldReturnBuiltProblemDetail() {
			var expectedProblemDetail = ProblemDetail.forStatusAndDetail(HttpStatus.UNPROCESSABLE_ENTITY,
					exceptionMessage);
			when(problemDetailMapper.fromConstraintViolationException(exception)).thenReturn(expectedProblemDetail);

			var problemDetail = handleException();

			assertThat(problemDetail).isEqualTo(expectedProblemDetail);
		}

		private ProblemDetail handleException() {
			return exceptionController.handleConstraintViolationException(exception);
		}
	}

	@SneakyThrows
	private String getDetailFromResponseContent(ResultActions resultActions) {
		return JsonPath.read(resultActions.andReturn().getResponse().getContentAsString(), "$.detail");
	}
}

@RestController
@RequestMapping(TestErrorController.BASE_PATH)
class TestErrorController {

	public final static String BASE_PATH = "/api/test-error";
	static final String ERROR_MESSAGE = "error message";

	@GetMapping("/runtime")
	String throwRuntimeException() throws Exception {
		throw new RuntimeException(ERROR_MESSAGE);
	}

	@GetMapping("/access-denied")
	String throwAccessException() throws Exception {
		throw new AccessDeniedException(ERROR_MESSAGE);
	}

	@GetMapping("/resource-not-found")
	String throwResourceNotFoundException() throws Exception {
		throw new ResourceNotFoundException(ERROR_MESSAGE);
	}

	@GetMapping("/functional")
	String throwFunctionalExceptionException() throws Exception {
		throw new FunctionalException(() -> ERROR_MESSAGE);
	}

	@GetMapping("/technical")
	String throwTechnicalExceptionException() throws Exception {
		throw new TechnicalException(ERROR_MESSAGE);
	}

}