/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.common.user;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Set;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.User;

import com.thedeanda.lorem.LoremIpsum;

class CurrentUserHelperTest {
	@DisplayName("Has role")
	@Nested
	class TestHasRole {
		@Mock
		private Authentication mockAuthentication;
		@Mock
		private User mockPrincipal;

		private final String role = LoremIpsum.getInstance().getWords(1);

		@Test
		void shouldCallHasAnyRole() {
			try (MockedStatic<CurrentUserHelper> mockUserHelper = mockStatic(CurrentUserHelper.class, CALLS_REAL_METHODS)) {
				mockUserHelper.when(() -> CurrentUserHelper.hasAnyRole(any())).thenReturn(false);

				CurrentUserHelper.hasRole(role);

				mockUserHelper.verify(() -> CurrentUserHelper.hasAnyRole(Set.of(role)));
			}
		}

		@ParameterizedTest
		@ValueSource(booleans = { false, true })
		void shouldReturnValue(boolean hasRole) {
			try (MockedStatic<CurrentUserHelper> mockUserHelper = mockStatic(CurrentUserHelper.class, CALLS_REAL_METHODS)) {
				mockUserHelper.when(() -> CurrentUserHelper.hasAnyRole(any())).thenReturn(hasRole);

				var result = CurrentUserHelper.hasRole(UserRole.ADMIN_ADMIN);

				assertThat(result).isEqualTo(hasRole);
			}
		}
	}

	@Nested
	class TestHasAnyRole {
		@Mock
		private Authentication mockAuthentication;
		@Mock
		private User mockPrincipal;

		private final Set<String> roles = Set.of(LoremIpsum.getInstance().getWords(1));

		@Test
		void shouldReturnFalseOnMissingAuthentication() {
			try (MockedStatic<CurrentUserHelper> mockUserHelper = mockStatic(CurrentUserHelper.class)) {
				mockUserHelper.when(CurrentUserHelper::getAuthentication).thenReturn(null);

				boolean hasRole = CurrentUserHelper.hasAnyRole(roles);

				assertThat(hasRole).isFalse();
			}
		}

		@Test
		void shouldReturnFalseOnMissingPrincipal() {
			when(mockAuthentication.getPrincipal()).thenReturn(null);
			try (MockedStatic<CurrentUserHelper> mockUserHelper = mockStatic(CurrentUserHelper.class, CALLS_REAL_METHODS)) {
				mockUserHelper.when(CurrentUserHelper::getAuthentication).thenReturn(mockAuthentication);

				boolean hasRole = CurrentUserHelper.hasAnyRole(roles);

				assertThat(hasRole).isFalse();
			}
		}

		@Test
		void shouldCallContainsAnyRole() {
			when(mockAuthentication.getPrincipal()).thenReturn(mockPrincipal);
			try (MockedStatic<CurrentUserHelper> mockUserHelper = mockStatic(CurrentUserHelper.class, CALLS_REAL_METHODS)) {
				mockUserHelper.when(CurrentUserHelper::getAuthentication).thenReturn(mockAuthentication);

				CurrentUserHelper.hasAnyRole(roles);

				mockUserHelper.verify(() -> CurrentUserHelper.containsAnyRole(mockAuthentication.getAuthorities(), roles));
			}
		}

		@ParameterizedTest
		@ValueSource(booleans = { false, true })
		void shouldReturnValue(boolean hasRole) {
			when(mockAuthentication.getPrincipal()).thenReturn(mockPrincipal);
			try (MockedStatic<CurrentUserHelper> mockUserHelper = mockStatic(CurrentUserHelper.class, CALLS_REAL_METHODS)) {
				mockUserHelper.when(CurrentUserHelper::getAuthentication).thenReturn(mockAuthentication);
				mockUserHelper.when(() -> CurrentUserHelper.containsAnyRole(any(), any())).thenReturn(hasRole);

				var result = CurrentUserHelper.hasAnyRole(roles);

				assertThat(result).isEqualTo(hasRole);
			}
		}
	}

	@Nested
	class TestContainsAnyRole {
		@Test
		void shouldReturnFalseIfAuthoritiesIsNull() {
			boolean containsRole = CurrentUserHelper.containsAnyRole(null, Set.of(LoremIpsum.getInstance().getWords(1)));

			assertThat(containsRole).isFalse();
		}

		@Test
		void shouldFalseOnNoMatchingRole() {
			var userRole = LoremIpsum.getInstance().getWords(1);
			var rolesToCheck = Set.of(LoremIpsum.getInstance().getWords(1), LoremIpsum.getInstance().getWords(1));
			List<GrantedAuthority> authorities = List.of(
					new SimpleGrantedAuthority(CurrentUserHelper.ROLE_PREFIX + userRole));

			boolean containsRole = CurrentUserHelper.containsAnyRole(authorities, rolesToCheck);

			assertThat(containsRole).isFalse();
		}

		@Test
		void shouldReturnTrueOnMatchingRole() {
			var userRole = LoremIpsum.getInstance().getWords(1);
			var rolesToCheck = Set.of(LoremIpsum.getInstance().getWords(1), userRole);
			List<GrantedAuthority> authorities = List.of(
					new SimpleGrantedAuthority(CurrentUserHelper.ROLE_PREFIX + userRole));

			boolean containsRole = CurrentUserHelper.containsAnyRole(authorities, rolesToCheck);

			assertThat(containsRole).isTrue();
		}
	}

	@DisplayName("Add Role Prefix If Missing")
	@Nested
	class TestAddRolePrefixIfMissing {

		@Test
		void shouldAddPrefixIfMissing() {
			var roleWithoutPrefix = UserRole.ADMIN_ADMIN;

			var role = CurrentUserHelper.addRolePrefixIfMissing(roleWithoutPrefix);

			assertThat(role).isEqualTo(CurrentUserHelper.ROLE_PREFIX + UserRole.ADMIN_ADMIN);
		}

		@Test
		void shouldReturnRoleIfPrefixAlreadyExists() {
			var roleWithPrefix = CurrentUserHelper.ROLE_PREFIX + UserRole.ADMIN_ADMIN;

			var role = CurrentUserHelper.addRolePrefixIfMissing(roleWithPrefix);

			assertThat(role).isEqualTo(roleWithPrefix);
		}

		@Test
		void shouldReturnNullIfPassingNull() {
			var role = CurrentUserHelper.addRolePrefixIfMissing(null);

			assertThat(role).isNull();
		}
	}

	@DisplayName("Get authentication")
	@Nested
	class TestGetAuthentication {
		@Mock
		private final SecurityContext mockSecurityContext = mock(SecurityContext.class);

		@Test
		void shouldThrowIfNoAuthenticatedUser() {
			when(mockSecurityContext.getAuthentication()).thenReturn(null);

			try (MockedStatic<SecurityContextHolder> contextHolder = mockStatic(SecurityContextHolder.class)) {
				contextHolder.when(SecurityContextHolder::getContext).thenReturn(mockSecurityContext);

				assertThatIllegalStateException()
						.isThrownBy(CurrentUserHelper::getAuthentication)
						.withMessage("No authenticated User found");
			}
		}

		@Test
		void shouldPassAuthentication() {
			Authentication mockAuthentication = mock(Authentication.class);
			when(mockSecurityContext.getAuthentication()).thenReturn(mockAuthentication);

			try (MockedStatic<SecurityContextHolder> contextHolder = mockStatic(SecurityContextHolder.class)) {
				contextHolder.when(SecurityContextHolder::getContext).thenReturn(mockSecurityContext);

				Authentication authentication = CurrentUserHelper.getAuthentication();

				assertThat(authentication).isSameAs(mockAuthentication);
			}
		}
	}
}
