/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.environment;

import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import de.ozgcloud.admin.RootController;
import de.ozgcloud.admin.common.FeatureToggleProperties;
import lombok.SneakyThrows;

class EnvironmentControllerTest {

	@Spy
	@InjectMocks
	private EnvironmentController controller;

	@Mock
	private ProductionProperties environmentProperties;
	@Mock
	private OAuth2Properties oAuth2Properties;
	@Mock
	private FeatureToggleProperties featureToggleProperties;
	@Mock
	private FeaturesMapper featuresMapper;

	private MockMvc mockMvc;

	@BeforeEach
	void mock() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
	}

	@Nested
	class TestGetEnvironment {
		@SneakyThrows
		@Test
		void shouldReturnOk() {
			when(environmentProperties.isProduction()).thenReturn(true);

			var response = doRequest();

			response.andExpect(status().isOk());
		}

		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		@SneakyThrows
		void shouldContainProduction(boolean production) {
			when(environmentProperties.isProduction()).thenReturn(production);

			var response = doRequest();

			response.andExpect(jsonPath("$.production").value(production));
		}

		@Test
		@SneakyThrows
		void shouldContainURLToRoot() {
			var response = doRequest();

			response.andExpect(jsonPath("$.remoteHost").value("http://localhost" + RootController.PATH));
		}

		@SneakyThrows
		@Test
		void shouldContainAuthServerUrl() {
			var oAuthServerUrl = "dummyOAuthUrl";
			when(oAuth2Properties.getAuthServerUrl()).thenReturn(oAuthServerUrl);

			var response = doRequest();

			response.andExpect(jsonPath("$.authServer").value(oAuthServerUrl));
		}

		@SneakyThrows
		@Test
		void shouldContainRealm() {
			var realm = "dummyRealm";
			when(oAuth2Properties.getRealm()).thenReturn(realm);

			var response = doRequest();

			response.andExpect(jsonPath("$.realm").value(realm));
		}

		@SneakyThrows
		@Test
		void shouldContainClientId() {
			var clientId = "dummdyClientId";
			when(oAuth2Properties.getResource()).thenReturn(clientId);

			var response = doRequest();

			response.andExpect(jsonPath("$.clientId").value(clientId));
		}

		@Test
		void shouldCallFeaturesMapper() {
			doRequest();

			verify(featuresMapper).fromFeatureToggleProperties(featureToggleProperties);
		}

		@SneakyThrows
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldContainPostfach(boolean postfach) {
			when(featuresMapper.fromFeatureToggleProperties(featureToggleProperties)).thenReturn(Features.builder().postfach(postfach).build());

			var response = doRequest();

			response.andExpect(jsonPath("$.features.postfach").value(postfach));
		}

		@SneakyThrows
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldContainBenutzerRollen(boolean benutzerRollen) {
			when(featuresMapper.fromFeatureToggleProperties(featureToggleProperties)).thenReturn(Features.builder().benutzerRollen(benutzerRollen).build());

			var response = doRequest();

			response.andExpect(jsonPath("$.features.benutzerRollen").value(benutzerRollen));
		}

		@SneakyThrows
		private ResultActions doRequest() {
			return mockMvc.perform(get(EnvironmentController.PATH));
		}

	}

}
