/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.keycloak;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.admin.Root;
import de.ozgcloud.admin.RootTestFactory;
import de.ozgcloud.admin.common.user.CurrentUserService;
import de.ozgcloud.admin.common.user.UserRole;

class KeyCloakRootProcessorTest {

	@InjectMocks
	@Spy
	private KeyCloakRootProcessor processor;
	@Mock
	private CurrentUserService currentUserService;
	@Mock
	private KeycloakApiProperties keycloakApiProperties;

	@Nested
	class TestProcess {
		private final String href = LoremIpsum.getInstance().getUrl();

		@Test
		void shouldCheckUserRole() {
			processModel();

			verify(currentUserService).hasRole(UserRole.ADMIN_ADMIN);
		}

		@Nested
		class TestOnAdminRole {

			@BeforeEach
			void givenHasAdminRole() {
				when(currentUserService.hasRole(anyString())).thenReturn(true);
				doReturn(href).when(processor).buildUsersHref();
			}

			@Test
			void shouldCallBuildUsersHref() {
				processModel();

				verify(processor).buildUsersHref();
			}

			@Test
			void shouldAddUsersLink() {
				var model = processModel();

				assertThat(model.getLink(KeyCloakRootProcessor.REL_USERS)).isNotEmpty();
			}

			@Test
			void shouldSetHref() {
				var model = processModel();

				assertThat(model.getLink(KeyCloakRootProcessor.REL_USERS))
						.get()
						.extracting(Link::getHref)
						.isEqualTo(href);
			}
		}

		@Nested
		class TestOnWrongUserRole {

			@BeforeEach
			void givenHasWrongRole() {
				when(currentUserService.hasRole(anyString())).thenReturn(false);
			}

			@Test
			void shouldNotAddAnyLinks() {
				var model = processModel();

				assertThat(model.getLinks()).isEmpty();
			}
		}

		private EntityModel<Root> processModel() {
			return processor.process(EntityModel.of(RootTestFactory.create()));
		}
	}

	@Nested
	class TestBuildUsersHref {
		private final String baseUrl = "https://keycloak.domain.de";
		private final String realm = LoremIpsum.getInstance().getWords(1);

		@BeforeEach
		void mockProperties() {
			when(keycloakApiProperties.getUrl()).thenReturn(baseUrl);
			when(keycloakApiProperties.getRealm()).thenReturn(realm);
		}

		@Test
		void shouldGetKeyCloakUrl() {
			processor.buildUsersHref();

			verify(keycloakApiProperties).getUrl();
		}

		@Test
		void shouldGetKeyCloakRealm() {
			processor.buildUsersHref();

			verify(keycloakApiProperties).getRealm();
		}

		@Test
		void shouldReturnApiEndpoint() {
			var uriString = processor.buildUsersHref();

			assertThat(uriString).isEqualTo(baseUrl + "/admin/realms/" + realm + "/users");
		}
	}
}