/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.keycloak;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.keycloak.representations.idm.GroupRepresentation;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

class KeycloakRemoteServiceTest {

	@Mock
	private KeycloakApiService apiService;
	@Mock
	private GroupMapper mapper;
	@Spy
	@InjectMocks
	private KeycloakRemoteService service;

	@Nested
	class TestGetGroupsWithOrganisationsEinheitId {

		private final List<GroupRepresentation> groupRepresentations = List.of(
			GroupRepresentationTestFactory.create("A"), GroupRepresentationTestFactory.create("B")
		);
		private final List<Group> mappedGroups = List.of(
			GroupTestFactory.createBuilder().name("A").build(), GroupTestFactory.createBuilder().name("B").build()
		);

		@BeforeEach
		void init() {
			when(apiService.getAllGroups()).thenReturn(groupRepresentations);
			when(mapper.fromGroupRepresentations(groupRepresentations)).thenReturn(mappedGroups);
		}

		@Test
		void shouldGetAllGroups() {
			service.getGroupsWithOrganisationsEinheitId();

			verify(apiService).getAllGroups();
		}

		@Test
		void shouldMapGroups() {
			service.getGroupsWithOrganisationsEinheitId();

			verify(mapper).fromGroupRepresentations(groupRepresentations);
		}

		@Test
		void shouldReturnMappedGroups() {
			var groups = service.getGroupsWithOrganisationsEinheitId();

			assertThat(groups).containsExactlyElementsOf(mappedGroups);
		}
	}

	@Nested
	class TestAddGroup {

		private static final String ADDED_GROUP_ID = UUID.randomUUID().toString();

		private final AddGroupData addGroupData = AddGroupDataTestFactory.create();
		private final GroupRepresentation groupRepresentation = GroupRepresentationTestFactory.create();

		@BeforeEach
		void init() {
			when(mapper.toGroupRepresentation(addGroupData)).thenReturn(groupRepresentation);
			when(apiService.addGroup(groupRepresentation)).thenReturn(ADDED_GROUP_ID);
		}

		@Test
		void shouldMapToGroupRepresentation() {
			callService();

			verify(mapper).toGroupRepresentation(addGroupData);
		}

		@Test
		void shouldAddGroupInKeycloak() {
			callService();

			verify(apiService).addGroup(groupRepresentation);
		}

		@Test
		void shouldReturnIdOfAddedGroup() {
			var id = callService();

			assertThat(id).isEqualTo(ADDED_GROUP_ID);
		}

		private String callService() {
			 return service.addGroup(addGroupData);
		}
	}
}
