/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.organisationseinheit;

import static de.ozgcloud.admin.organisationseinheit.OrganisationsEinheitModelAssembler.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Collection;
import java.util.List;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Spy;
import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.IanaLinkRelations;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.LinkRelation;
import org.springframework.hateoas.mediatype.hal.HalModelBuilder;
import org.springframework.web.util.UriTemplate;

class OrganisationsEinheitModelAssemblerTest {

	@Spy
	@InjectMocks
	private OrganisationsEinheitModelAssembler assembler;

	@Nested
	class TestToModel {

		private final OrganisationsEinheit organisationsEinheit = OrganisationsEinheitTestFactory.createBuilder().parentId(null).build();

		@BeforeEach
		void setUp() {
			doNothing().when(assembler).embedChildOrganisationsEinheiten(eq(organisationsEinheit), any());
		}

		@Test
		void shouldAddSelfLink() {
			var model = assembler.toModel(organisationsEinheit);

			assertThat(model.getLink(IanaLinkRelations.SELF))
					.isNotEmpty()
					.get()
					.extracting(Link::getHref)
					.isEqualTo(new UriTemplate(OrganisationsEinheitController.PATH + "/{id}").expand(OrganisationsEinheitTestFactory.ID).toString());
		}

		@Test
		void shouldReturnModel() {
			var model = assembler.toModel(organisationsEinheit);

			assertThat(model.getContent()).isEqualTo(organisationsEinheit);
		}

		@Test
		void shouldEmbedChildOrganisationsEinheiten() {
			assembler.toModel(organisationsEinheit);

			verify(assembler).embedChildOrganisationsEinheiten(eq(organisationsEinheit), any());
		}

		@Test
		void shouldAddEditLink() {
			var model = assembler.toModel(organisationsEinheit);

			assertThat(model.getLink(OrganisationsEinheitModelAssembler.REL_EDIT))
					.isNotEmpty()
					.get()
					.extracting(Link::getHref)
					.isEqualTo(new UriTemplate(OrganisationsEinheitController.PATH + "/{id}").expand(OrganisationsEinheitTestFactory.ID).toString());
		}
	}

	@Nested
	class TestEmbedChildOrganisationsEinheiten {

		private OrganisationsEinheit parentOrganisationsEinheit;
		private OrganisationsEinheit childOrganisationsEinheit;
		private CollectionModel<EntityModel<OrganisationsEinheit>> childrenCollectionModel;
		@Spy
		private final HalModelBuilder halModelBuilder = HalModelBuilder.emptyHalModel();
		@Captor
		private ArgumentCaptor<LinkRelation> linkRelationArgumentCaptor;
		@Captor
		private ArgumentCaptor<Collection<EntityModel<OrganisationsEinheit>>> collectionModelContentArgumentCaptor;

		@BeforeEach
		void setUp() {
			parentOrganisationsEinheit = OrganisationsEinheitTestFactory.createBuilder().parentId(null).build();
			childOrganisationsEinheit = OrganisationsEinheitTestFactory.createBuilder().id(UUID.randomUUID().toString())
					.parentId(parentOrganisationsEinheit.getId()).build();
			parentOrganisationsEinheit = parentOrganisationsEinheit.toBuilder().child(childOrganisationsEinheit).build();
			childrenCollectionModel = CollectionModel.of(List.of(EntityModel.of(childOrganisationsEinheit)));
			doReturn(childrenCollectionModel).when(assembler)
					.toChildrenCollectionModel(parentOrganisationsEinheit.getId(), parentOrganisationsEinheit.getChildren());
		}

		@Test
		void shouldBuildChildrenCollectionModel() {
			assembler.embedChildOrganisationsEinheiten(parentOrganisationsEinheit, halModelBuilder);

			verify(assembler).toChildrenCollectionModel(parentOrganisationsEinheit.getId(), parentOrganisationsEinheit.getChildren());
		}

		@Test
		void shouldEmbedChildrenModels() {
			assembler.embedChildOrganisationsEinheiten(parentOrganisationsEinheit, halModelBuilder);

			verify(halModelBuilder).embed(collectionModelContentArgumentCaptor.capture(), linkRelationArgumentCaptor.capture());
			assertThat(collectionModelContentArgumentCaptor.getValue()).usingRecursiveComparison().isEqualTo(childrenCollectionModel.getContent());
			assertThat(linkRelationArgumentCaptor.getValue().toString()).isEqualTo(REL_CHILD_ORGANISATIONS_EINHEITEN);
		}

		@Test
		void shouldHaveChildListLinkRelation() {
			assembler.embedChildOrganisationsEinheiten(parentOrganisationsEinheit, halModelBuilder);

			assertThat(halModelBuilder.build().getLink(REL_CHILD_ORGANISATIONS_EINHEITEN))
					.isNotEmpty()
					.get()
					.extracting(Link::getHref)
					.isEqualTo(OrganisationsEinheitController.PATH + "/" + parentOrganisationsEinheit.getId() + "/children");
		}
	}

	@Nested
	class TestToCollectionModel {

		private final OrganisationsEinheit organisationsEinheit = OrganisationsEinheitTestFactory.create();

		@BeforeEach
		void setUp() {
			doReturn(EntityModel.of(organisationsEinheit)).when(assembler).toModel(organisationsEinheit);
		}

		@Test
		void shouldAddSelfLink() {
			var collectionModel = assembler.toCollectionModel(List.of(organisationsEinheit));

			assertThat(collectionModel.getLink(IanaLinkRelations.SELF))
					.isNotEmpty()
					.get()
					.extracting(Link::getHref)
					.isEqualTo(OrganisationsEinheitController.PATH);
		}

		@Test
		void shouldBuildModels() {
			assembler.toCollectionModel(List.of(organisationsEinheit));

			verify(assembler).toModel(organisationsEinheit);
		}

		@Test
		void shouldHaveModels() {
			var collectionModel = assembler.toCollectionModel(List.of(organisationsEinheit));

			assertThat(collectionModel.getContent()).extracting(EntityModel::getContent).containsExactly(organisationsEinheit);
		}
	}

	@Nested
	class ToChildrenCollectionModel {

		private final OrganisationsEinheit parentOrganisationsEinheit = OrganisationsEinheitTestFactory.createBuilder().parentId(null).build();
		private final OrganisationsEinheit childOrganisationsEinheit = OrganisationsEinheitTestFactory.createBuilder()
				.id(UUID.randomUUID().toString()).build();

		@BeforeEach
		void setUp() {
			doReturn(EntityModel.of(childOrganisationsEinheit)).when(assembler).toModel(childOrganisationsEinheit);
		}

		@Test
		void shouldAddSelfLink() {
			var collectionModel = assembler.toChildrenCollectionModel(parentOrganisationsEinheit.getId(), List.of(childOrganisationsEinheit));

			assertThat(collectionModel.getLink(IanaLinkRelations.SELF))
					.isNotEmpty()
					.get()
					.extracting(Link::getHref)
					.isEqualTo(OrganisationsEinheitController.PATH + "/" + parentOrganisationsEinheit.getId() + "/children");
		}

		@Test
		void shouldBuildModels() {
			assembler.toChildrenCollectionModel(parentOrganisationsEinheit.getId(), List.of(childOrganisationsEinheit));

			verify(assembler).toModel(childOrganisationsEinheit);
		}

		@Test
		void shouldHaveModels() {
			var collectionModel = assembler.toChildrenCollectionModel(parentOrganisationsEinheit.getId(), List.of(childOrganisationsEinheit));

			assertThat(collectionModel.getContent()).extracting(EntityModel::getContent).containsExactly(childOrganisationsEinheit);
		}

	}
}