/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.organisationseinheit;

import static org.assertj.core.api.Assertions.*;
import static org.assertj.core.groups.Tuple.tuple;

import java.time.Instant;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.EnableAutoConfiguration;
import org.springframework.boot.test.autoconfigure.data.mongo.DataMongoTest;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.ContextConfiguration;

import de.ozgcloud.common.test.DbInitializer;

@DataMongoTest
@ContextConfiguration(initializers = { DbInitializer.class }, classes = { OrganisationsEinheitRepository.class, MongoOperations.class })
@ActiveProfiles({ "itcase", "with_db" })
@EnableAutoConfiguration
class OrganisationsEinheitRepositoryITCase {

	@Autowired
	private OrganisationsEinheitRepository repository;

	@Autowired
	private MongoOperations operations;

	@BeforeEach
	void clearDatabase() {
		operations.dropCollection(OrganisationsEinheit.class);
	}

	@Nested
	class TestFindAllNotDeleted {

		private static final String DELETED_ORGANISATIONS_EINHEIT_ID_1 = "1";
		private static final String DELETED_ORGANISATIONS_EINHEIT_ID_2 = "2";
		private static final String SYNCED_ORGANISATIONS_EINHEIT_ID_1 = "3";
		private static final String SYNCED_ORGANISATIONS_EINHEIT_ID_2 = "4";
		private static final String ADDED_FROM_ZUFI_ORGANISATIONS_EINHEIT_ID_1 = "5";

		private final OrganisationsEinheit deleted1 = OrganisationsEinheitTestFactory.createBuilder()
				.id(null)
				.organisationsEinheitId(DELETED_ORGANISATIONS_EINHEIT_ID_1)
				.syncResult(SyncResult.DELETED)
				.build();
		private final OrganisationsEinheit deleted2 = OrganisationsEinheitTestFactory.createBuilder()
				.id(null)
				.organisationsEinheitId(DELETED_ORGANISATIONS_EINHEIT_ID_2)
				.syncResult(SyncResult.DELETED)
				.build();
		private final OrganisationsEinheit synced1 = OrganisationsEinheitTestFactory.createBuilder()
				.id(null)
				.organisationsEinheitId(SYNCED_ORGANISATIONS_EINHEIT_ID_1)
				.build();
		private final OrganisationsEinheit synced2 = OrganisationsEinheitTestFactory.createBuilder()
				.id(null)
				.organisationsEinheitId(SYNCED_ORGANISATIONS_EINHEIT_ID_2)
				.build();
		private final OrganisationsEinheit addedFromZufi = OrganisationsEinheitTestFactory.createBuilder()
				.id(null)
				.syncResult(null)
				.organisationsEinheitId(ADDED_FROM_ZUFI_ORGANISATIONS_EINHEIT_ID_1)
				.build();

		@BeforeEach
		void setUp() {
			operations.save(deleted1);
			operations.save(deleted2);
			operations.save(synced1);
			operations.save(synced2);
			operations.save(addedFromZufi);
		}

		@Test
		void shouldReturnNotDeleted() {
			var allNotDeleted = repository.findAllNotDeleted();

			assertThat(allNotDeleted).extracting(
					OrganisationsEinheit::getOrganisationsEinheitId,
					OrganisationsEinheit::getSyncResult
			).containsExactly(
					tuple(SYNCED_ORGANISATIONS_EINHEIT_ID_1, OrganisationsEinheitTestFactory.SYNC_RESULT),
					tuple(SYNCED_ORGANISATIONS_EINHEIT_ID_2, OrganisationsEinheitTestFactory.SYNC_RESULT),
					tuple(ADDED_FROM_ZUFI_ORGANISATIONS_EINHEIT_ID_1, null)
			);
		}
	}

	@Nested
	class TestFindSyncedByKeycloakId {

		@Test
		void shouldReturnEmptyOnEmptyDatabase() {
			var organisationsEinheit = repository.findSyncedByKeycloakId(OrganisationsEinheitTestFactory.KEYCLOAK_ID);

			assertThat(organisationsEinheit).isEmpty();
		}

		@Test
		void shouldReturnEmptyOnNotFoundKeycloakId() {
			operations.save(OrganisationsEinheitTestFactory.createBuilder().id(null).build());

			var organisationsEinheit = repository.findSyncedByKeycloakId("not_found");

			assertThat(organisationsEinheit).isEmpty();
		}

		@Test
		void shouldReturnEmptyOnSyncResultNull() {
			operations.save(OrganisationsEinheitTestFactory.createBuilder().id(null).syncResult(null).build());

			var organisationsEinheit = repository.findSyncedByKeycloakId(OrganisationsEinheitTestFactory.KEYCLOAK_ID);

			assertThat(organisationsEinheit).isEmpty();
		}

		@Test
		void shouldFind() {
			operations.save(OrganisationsEinheitTestFactory.createBuilder().id(null).build());
			operations.save(OrganisationsEinheitTestFactory.createBuilder().id(null).keycloakId(UUID.randomUUID().toString()).build());

			var organisationsEinheit = repository.findSyncedByKeycloakId(OrganisationsEinheitTestFactory.KEYCLOAK_ID);

			assertThat(organisationsEinheit)
					.isNotEmpty()
					.get()
					.extracting(OrganisationsEinheit::getKeycloakId)
					.isEqualTo(OrganisationsEinheitTestFactory.KEYCLOAK_ID);
		}
	}

	@Nested
	class TestSetUnsyncedAsDeleted {

		private static final String DELETED_ORGANISATIONS_EINHEIT_ID_1 = "1";
		private static final String DELETED_ORGANISATIONS_EINHEIT_ID_2 = "2";
		private static final String SYNCED_ORGANISATIONS_EINHEIT_ID_1 = "3";
		private static final String SYNCED_ORGANISATIONS_EINHEIT_ID_2 = "4";

		private final OrganisationsEinheit deleted1 = OrganisationsEinheitTestFactory.createBuilder()
				.id(null)
				.organisationsEinheitId(DELETED_ORGANISATIONS_EINHEIT_ID_1)
				.lastSyncTimestamp(OrganisationsEinheitTestFactory.LAST_SYNC_UPDATE - 1)
				.build();
		private final OrganisationsEinheit deleted2 = OrganisationsEinheitTestFactory.createBuilder()
				.id(null)
				.organisationsEinheitId(DELETED_ORGANISATIONS_EINHEIT_ID_2)
				.lastSyncTimestamp(OrganisationsEinheitTestFactory.LAST_SYNC_UPDATE - 2)
				.build();
		private final OrganisationsEinheit synced1 = OrganisationsEinheitTestFactory.createBuilder()
				.id(null)
				.organisationsEinheitId(SYNCED_ORGANISATIONS_EINHEIT_ID_1)
				.build();
		private final OrganisationsEinheit synced2 = OrganisationsEinheitTestFactory.createBuilder()
				.id(null)
				.organisationsEinheitId(SYNCED_ORGANISATIONS_EINHEIT_ID_2)
				.build();

		@BeforeEach
		void setUp() {
			operations.save(deleted1);
			operations.save(deleted2);
			operations.save(synced1);
			operations.save(synced2);
		}

		@Test
		void shouldSetUnsyncedAsDeleted() {
			repository.setUnsyncedAsDeleted(OrganisationsEinheitTestFactory.LAST_SYNC_UPDATE);

			var all = repository.findAll();

			assertThat(all).extracting(
					OrganisationsEinheit::getOrganisationsEinheitId,
					OrganisationsEinheit::getSyncResult
			).containsExactly(
					tuple(DELETED_ORGANISATIONS_EINHEIT_ID_1, SyncResult.DELETED),
					tuple(DELETED_ORGANISATIONS_EINHEIT_ID_2, SyncResult.DELETED),
					tuple(SYNCED_ORGANISATIONS_EINHEIT_ID_1, OrganisationsEinheitTestFactory.SYNC_RESULT),
					tuple(SYNCED_ORGANISATIONS_EINHEIT_ID_2, OrganisationsEinheitTestFactory.SYNC_RESULT)
			);
		}

		@Test
		void shouldSetAllAsDeleted() {
			repository.setUnsyncedAsDeleted(Instant.now().toEpochMilli());

			var all = repository.findAll();

			assertThat(all).extracting(
					OrganisationsEinheit::getOrganisationsEinheitId,
					OrganisationsEinheit::getSyncResult
			).containsExactly(
					tuple(DELETED_ORGANISATIONS_EINHEIT_ID_1, SyncResult.DELETED),
					tuple(DELETED_ORGANISATIONS_EINHEIT_ID_2, SyncResult.DELETED),
					tuple(SYNCED_ORGANISATIONS_EINHEIT_ID_1, SyncResult.DELETED),
					tuple(SYNCED_ORGANISATIONS_EINHEIT_ID_2, SyncResult.DELETED)
			);
		}
	}

	@Nested
	class TestFindAllWithoutSyncResult {

		private static final String SYNCED_ID_1 = "1";
		private static final String WITHOUT_SYNC_RESULT_ID_1 = "2";
		private static final String WITHOUT_SYNC_RESULT_ID_2 = "3";

		private final OrganisationsEinheit synced1 = OrganisationsEinheitTestFactory.createBuilder()
				.id(SYNCED_ID_1)
				.build();
		private final OrganisationsEinheit withoutSyncResult1 = OrganisationsEinheitTestFactory.createBuilder()
				.id(WITHOUT_SYNC_RESULT_ID_1)
				.syncResult(null)
				.build();
		private final OrganisationsEinheit withoutSyncResult2 = OrganisationsEinheitTestFactory.createBuilder()
				.id(WITHOUT_SYNC_RESULT_ID_2)
				.parentId(SYNCED_ID_1)
				.syncResult(null)
				.build();

		@BeforeEach
		void setUp() {
			operations.save(synced1);
			operations.save(withoutSyncResult1);
			operations.save(withoutSyncResult2);
		}

		@Test
		void shouldFindAllWithoutSyncResult() {
			var allWithoutSyncResult = repository.findAllWithoutSyncResult();

			assertThat(allWithoutSyncResult).extracting(OrganisationsEinheit::getId)
					.containsExactlyInAnyOrder(WITHOUT_SYNC_RESULT_ID_1, WITHOUT_SYNC_RESULT_ID_2);
		}
	}

	@Nested
	class TestFindChildren {

		private final OrganisationsEinheit root1 = OrganisationsEinheitTestFactory.createBuilder().parentId(null).build();
		private final OrganisationsEinheit root2 = OrganisationsEinheitTestFactory.createBuilder()
				.id(UUID.randomUUID().toString())
				.parentId(null)
				.build();
		private final OrganisationsEinheit child1 = OrganisationsEinheitTestFactory.createBuilder()
				.id(UUID.randomUUID().toString())
				.parentId(root1.getId())
				.build();
		private final OrganisationsEinheit child2 = OrganisationsEinheitTestFactory.createBuilder()
				.id(UUID.randomUUID().toString())
				.parentId(root1.getId())
				.build();

		private final OrganisationsEinheit child3 = OrganisationsEinheitTestFactory.createBuilder()
				.id(UUID.randomUUID().toString())
				.parentId(root2.getId())
				.build();

		@BeforeEach
		void setUp() {
			operations.save(root1);
			operations.save(root2);
			operations.save(child1);
			operations.save(child2);
			operations.save(child3);
		}

		@Test
		void shouldFind() {
			var found = repository.findChildren(root1.getId());

			assertThat(found).extracting(OrganisationsEinheit::getId).containsOnly(child1.getId(), child2.getId());
		}

		@Test
		void shouldNotFind() {
			var found = repository.findChildren(UUID.randomUUID().toString());

			assertThat(found).isEmpty();

		}
	}
}