/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.organisationseinheit;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.admin.Root;
import de.ozgcloud.admin.RootTestFactory;
import de.ozgcloud.admin.common.user.CurrentUserService;
import de.ozgcloud.admin.common.user.UserRole;

class OrganisationsEinheitRootProcessorTest {

	@InjectMocks
	private OrganisationsEinheitRootProcessor organisationsEinheitRootProcessor;
	@Mock
	private OrganisationEinheitProperties organisationsEinheitProperties;
	@Mock
	private CurrentUserService currentUserService;

	@Nested
	class TestProcess {

		private final String zufiSearchUri = LoremIpsum.getInstance().getUrl() + "?searchBy={searchBy}";

		@Test
		void shouldCheckUserRole() {
			processModel();

			verify(currentUserService).hasRole(UserRole.ADMIN_ADMIN);
		}

		@Nested
		class TestOnWrongUserRole {

			@BeforeEach
			void givenHasWrongRole() {
				when(currentUserService.hasRole(anyString())).thenReturn(false);
			}

			@Test
			void shouldNotAddAnyLinks() {
				var model = processModel();

				assertThat(model.getLinks()).isEmpty();
			}
		}

		@Nested
		class TestOnAdminRole {

			@BeforeEach
			void givenHasAdminRole() {
				when(currentUserService.hasRole(anyString())).thenReturn(true);
				when(organisationsEinheitProperties.getZufiSearchUri()).thenReturn(zufiSearchUri);
			}

			@Nested
			class TestOrganisationsEinheitenLinkRelation {

				@Test
				void shouldExist() {
					var model = processModel();

					assertThat(model.getLink(OrganisationsEinheitRootProcessor.REL_ORGANISATION_EINHEITEN)).isNotEmpty();
				}

				@Test
				void shouldHaveHref() {
					var model = processModel();

					assertThat(model.getLink(OrganisationsEinheitRootProcessor.REL_ORGANISATION_EINHEITEN))
							.get()
							.extracting(Link::getHref)
							.isEqualTo(OrganisationsEinheitController.PATH);
				}
			}

			@Nested
			class TestSearchOrganisationsEinheitenLink {

				@Test
				void shouldExistsIfFeatureEnabled() {
					var model = processModel();

					assertThat(model.getLink(OrganisationsEinheitRootProcessor.REL_SEARCH_ORGANISATION_EINHEIT)).isNotEmpty();
				}

				@Test
				void shouldGetZufiSearchUri() {
					processModel();

					verify(organisationsEinheitProperties).getZufiSearchUri();

				}

				@Test
				void shouldHaveHref() {
					var model = processModel();

					assertThat(model.getLink(OrganisationsEinheitRootProcessor.REL_SEARCH_ORGANISATION_EINHEIT))
							.get()
							.extracting(Link::getHref)
							.isEqualTo(zufiSearchUri);
				}

				@Test
				void shouldBeTemplated() {
					var model = processModel();

					assertThat(model.getLink(OrganisationsEinheitRootProcessor.REL_SEARCH_ORGANISATION_EINHEIT))
							.get()
							.extracting(Link::isTemplated)
							.isEqualTo(true);
				}
			}
		}

		private EntityModel<Root> processModel() {
			return organisationsEinheitRootProcessor.process(EntityModel.of(RootTestFactory.create()));
		}
	}
}