/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.organisationseinheit;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.time.Instant;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.EnableAutoConfiguration;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.data.mongodb.core.MongoOperations;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.admin.keycloak.GroupTestFactory;
import de.ozgcloud.admin.keycloak.KeycloakRemoteService;
import de.ozgcloud.common.test.DataITCase;

@DataITCase
@EnableAutoConfiguration
class SyncServiceITCase {

	@Autowired
	private MongoOperations operations;

	@Autowired
	private SyncService service;

	@Autowired
	private OrganisationsEinheitRepository repository;

	@MockBean
	private KeycloakRemoteService keycloakRemoteService;

	@MockBean
	private OrganisationsEinheitRemoteService organisationsEinheitRemoteService;

	@BeforeEach
	void clearDatabase() {
		operations.dropCollection(OrganisationsEinheit.class);
	}

	@DisplayName("Organisationseinheit not found in PVOG")
	@Test
	void shouldSyncNotFoundInPvog() {
		var syncTimestamp = Instant.now().toEpochMilli();
		var group = GroupTestFactory.createBuilder().clearSubGroups().build();
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of(group));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(List.of());

		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp);

		var synced = repository.findAll();

		assertThat(synced).
				hasSize(1)
				.first()
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						group.getId(),
						group.getName(),
						group.getOrganisationsEinheitId(),
						SyncResult.NOT_FOUND_IN_PVOG,
						null,
						null,
						Collections.emptyList(),
						syncTimestamp);
	}

	@DisplayName("Organisationseinheit found in PVOG but group name is different")
	@Test
	void shouldSyncNameMismatch() {
		var syncTimestamp = Instant.now().toEpochMilli();
		var group = GroupTestFactory.createBuilder().clearSubGroups().build();
		var pvogOranigastionsEinheit = OrganisationsEinheitTestFactory.createBuilder().organisationsEinheitId(group.getOrganisationsEinheitId())
				.build();
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of(group));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogOranigastionsEinheit));

		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp);

		var synced = repository.findAll();

		assertThat(synced).
				hasSize(1)
				.first()
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						group.getId(),
						pvogOranigastionsEinheit.getName(),
						group.getOrganisationsEinheitId(),
						SyncResult.NAME_MISMATCH,
						pvogOranigastionsEinheit.getId(),
						null,
						Collections.emptyList(),
						syncTimestamp);
	}

	@DisplayName("Multiple Organisationseinheiten for same group found in PVOG")
	@Test
	void shouldSyncOrganisationsEinheitNotUnique() {
		var syncTimestamp = Instant.now().toEpochMilli();
		var group = GroupTestFactory.createBuilder().clearSubGroups().build();
		var pvogOranigastionsEinheit1 = OrganisationsEinheitTestFactory.createBuilder().organisationsEinheitId(group.getOrganisationsEinheitId())
				.build();
		var pvogOranigastionsEinheit2 = OrganisationsEinheitTestFactory.createBuilder().organisationsEinheitId(group.getOrganisationsEinheitId())
				.build();
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of(group));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogOranigastionsEinheit1, pvogOranigastionsEinheit2));

		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp);

		var synced = repository.findAll();

		assertThat(synced).
				hasSize(1)
				.first()
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						group.getId(),
						group.getName(),
						group.getOrganisationsEinheitId(),
						SyncResult.ORGANISATIONSEINHEIT_ID_NOT_UNIQUE,
						null,
						null,
						Collections.emptyList(),
						syncTimestamp);
	}

	@DisplayName("Organisationseinheit found in PVOG and have same data")
	@Test
	void shouldSyncOk() {
		var syncTimestamp = Instant.now().toEpochMilli();
		var group = GroupTestFactory.createBuilder().clearSubGroups().build();
		var pvogOranigastionsEinheit = OrganisationsEinheitTestFactory.createBuilder().organisationsEinheitId(group.getOrganisationsEinheitId())
				.name(group.getName()).build();
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of(group));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogOranigastionsEinheit));

		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp);

		var synced = repository.findAll();

		assertThat(synced).
				hasSize(1)
				.first()
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						group.getId(),
						group.getName(),
						group.getOrganisationsEinheitId(),
						SyncResult.OK,
						pvogOranigastionsEinheit.getId(),
						null,
						Collections.emptyList(),
						syncTimestamp);
	}

	@DisplayName("OrganisationseinheitId attribute of group changed")
	@Test
	void shouldSyncUpdateAttribute() {
		var syncTimestamp = Instant.now().toEpochMilli();
		var group = GroupTestFactory.createBuilder().clearSubGroups().build();
		var pvogOranigastionsEinheit = OrganisationsEinheitTestFactory.createBuilder().organisationsEinheitId(group.getOrganisationsEinheitId())
				.name(group.getName()).build();
		var newOrganisationsEinheitId = UUID.randomUUID().toString();
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of(group));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogOranigastionsEinheit));

		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp);
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(
				Stream.of(group.toBuilder().organisationsEinheitId(newOrganisationsEinheitId).build()));
		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp + 1000);

		var synced = repository.findAll();

		assertThat(synced).
				hasSize(1)
				.first()
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						group.getId(),
						group.getName(),
						newOrganisationsEinheitId,
						SyncResult.NOT_FOUND_IN_PVOG,
						null,
						null,
						Collections.emptyList(),
						syncTimestamp + 1000);
	}

	@DisplayName("Group deleted in Keycloak")
	@Test
	void shouldSyncDeleted() {
		var syncTimestamp = Instant.now().toEpochMilli();
		var group = GroupTestFactory.createBuilder().clearSubGroups().build();
		var pvogOranigastionsEinheit = OrganisationsEinheitTestFactory.createBuilder().organisationsEinheitId(group.getOrganisationsEinheitId())
				.name(group.getName()).build();
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of(group));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogOranigastionsEinheit));

		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp);
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of());
		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp + 1000);

		var synced = repository.findAll();

		assertThat(synced).
				hasSize(1)
				.first()
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						group.getId(),
						group.getName(),
						group.getOrganisationsEinheitId(),
						SyncResult.DELETED,
						pvogOranigastionsEinheit.getId(),
						null,
						Collections.emptyList(),
						syncTimestamp);
	}

	@DisplayName("Organisationseinheit for parent group and sub group not found in PVOG")
	@Test
	void shouldSyncParentNotFoundInPvogSubGroupNotFoundInPvog() {
		var syncTimestamp = Instant.now().toEpochMilli();
		var group = GroupTestFactory.createBuilder().build();
		var subGroup = group.getSubGroups().getFirst();
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of(group));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(List.of());

		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp);

		var parentSynced = repository.findSyncedByKeycloakId(group.getId()).get();
		var childSynced = repository.findSyncedByKeycloakId(subGroup.getId()).get();

		assertThat(parentSynced)
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						group.getId(),
						group.getName(),
						group.getOrganisationsEinheitId(),
						SyncResult.NOT_FOUND_IN_PVOG,
						null,
						null,
						Collections.emptyList(),
						syncTimestamp);
		assertThat(childSynced)
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						subGroup.getId(),
						subGroup.getName(),
						subGroup.getOrganisationsEinheitId(),
						SyncResult.NOT_FOUND_IN_PVOG,
						null,
						parentSynced.getId(),
						Collections.emptyList(),
						syncTimestamp);
	}

	@DisplayName("Organisationseinheit for parent group and sub group found in PVOG but have different names")
	@Test
	void shouldSyncParentNameMismatchSubGroupNameMismatch() {
		var syncTimestamp = Instant.now().toEpochMilli();
		var group = GroupTestFactory.createBuilder().build();
		var subGroup = group.getSubGroups().getFirst();
		var pvogOranigastionsEinheit = OrganisationsEinheitTestFactory.createBuilder().organisationsEinheitId(group.getOrganisationsEinheitId())
				.build();
		var pvogChildOranigastionsEinheit = OrganisationsEinheitTestFactory.createBuilder()
				.organisationsEinheitId(subGroup.getOrganisationsEinheitId()).name(
						LoremIpsum.getInstance().getWords(2))
				.build();
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of(group));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogOranigastionsEinheit));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.SUB_GROUP_ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogChildOranigastionsEinheit));

		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp);

		var parentSynced = repository.findSyncedByKeycloakId(group.getId()).get();
		var childSynced = repository.findSyncedByKeycloakId(subGroup.getId()).get();

		assertThat(parentSynced)
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						group.getId(),
						pvogOranigastionsEinheit.getName(),
						group.getOrganisationsEinheitId(),
						SyncResult.NAME_MISMATCH,
						pvogOranigastionsEinheit.getId(),
						null,
						Collections.emptyList(),
						syncTimestamp);
		assertThat(childSynced)
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						subGroup.getId(),
						pvogChildOranigastionsEinheit.getName(),
						subGroup.getOrganisationsEinheitId(),
						SyncResult.NAME_MISMATCH,
						pvogChildOranigastionsEinheit.getId(),
						parentSynced.getId(),
						Collections.emptyList(),
						syncTimestamp);
	}

	@DisplayName("Multiple Organisationseinheiten for parent group and sub group found in PVOG")
	@Test
	void shouldSyncParentAndChildOrganisationsEinheitNotUnique() {
		var syncTimestamp = Instant.now().toEpochMilli();
		var group = GroupTestFactory.createBuilder().build();
		var subGroup = group.getSubGroups().getFirst();
		var pvogOranigastionsEinheit1 = OrganisationsEinheitTestFactory.createBuilder().organisationsEinheitId(group.getOrganisationsEinheitId())
				.build();
		var pvogOranigastionsEinheit2 = OrganisationsEinheitTestFactory.createBuilder().organisationsEinheitId(group.getOrganisationsEinheitId())
				.build();
		var pvogChildOranigastionsEinheit1 = OrganisationsEinheitTestFactory.createBuilder()
				.organisationsEinheitId(subGroup.getOrganisationsEinheitId()).name(
						LoremIpsum.getInstance().getWords(2))
				.build();
		var pvogChildOranigastionsEinheit2 = OrganisationsEinheitTestFactory.createBuilder()
				.organisationsEinheitId(subGroup.getOrganisationsEinheitId()).name(
						LoremIpsum.getInstance().getWords(2))
				.build();
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of(group));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogOranigastionsEinheit1, pvogOranigastionsEinheit2));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.SUB_GROUP_ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogChildOranigastionsEinheit1, pvogChildOranigastionsEinheit2));

		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp);

		var parentSynced = repository.findSyncedByKeycloakId(group.getId()).get();
		var childSynced = repository.findSyncedByKeycloakId(subGroup.getId()).get();

		assertThat(parentSynced)
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						group.getId(),
						group.getName(),
						group.getOrganisationsEinheitId(),
						SyncResult.ORGANISATIONSEINHEIT_ID_NOT_UNIQUE,
						null,
						null,
						Collections.emptyList(),
						syncTimestamp);
		assertThat(childSynced)
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						subGroup.getId(),
						subGroup.getName(),
						subGroup.getOrganisationsEinheitId(),
						SyncResult.ORGANISATIONSEINHEIT_ID_NOT_UNIQUE,
						null,
						parentSynced.getId(),
						Collections.emptyList(),
						syncTimestamp);
	}

	@DisplayName("Organisationseinheit and child Organisationseinheit found in PVOG and have same data")
	@Test
	void shouldSyncOkGroupWithSubGroup() {
		var syncTimestamp = Instant.now().toEpochMilli();
		var group = GroupTestFactory.createBuilder().build();
		var subGroup = group.getSubGroups().getFirst();
		var pvogOranigastionsEinheit = OrganisationsEinheitTestFactory.createBuilder().organisationsEinheitId(group.getOrganisationsEinheitId())
				.name(group.getName()).build();
		var pvogChildOranigastionsEinheit = OrganisationsEinheitTestFactory.createBuilder()
				.organisationsEinheitId(subGroup.getOrganisationsEinheitId()).name(
						subGroup.getName())
				.build();
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of(group));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogOranigastionsEinheit));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.SUB_GROUP_ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogChildOranigastionsEinheit));

		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp);

		var parentSynced = repository.findSyncedByKeycloakId(group.getId()).get();
		var childSynced = repository.findSyncedByKeycloakId(subGroup.getId()).get();

		assertThat(parentSynced)
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						group.getId(),
						group.getName(),
						group.getOrganisationsEinheitId(),
						SyncResult.OK,
						pvogOranigastionsEinheit.getId(),
						null,
						Collections.emptyList(),
						syncTimestamp);
		assertThat(childSynced)
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						subGroup.getId(),
						subGroup.getName(),
						subGroup.getOrganisationsEinheitId(),
						SyncResult.OK,
						pvogChildOranigastionsEinheit.getId(),
						parentSynced.getId(),
						Collections.emptyList(),
						syncTimestamp);
	}

	@DisplayName("Group with sub group deleted in Keycloak")
	@Test
	void shouldSyncGroupAndSubGroupDeleted() {
		var syncTimestamp = Instant.now().toEpochMilli();
		var group = GroupTestFactory.createBuilder().build();
		var subGroup = group.getSubGroups().getFirst();
		var pvogOranigastionsEinheit = OrganisationsEinheitTestFactory.createBuilder().organisationsEinheitId(group.getOrganisationsEinheitId())
				.name(group.getName()).build();
		var pvogChildOranigastionsEinheit = OrganisationsEinheitTestFactory.createBuilder()
				.organisationsEinheitId(subGroup.getOrganisationsEinheitId()).name(
						subGroup.getName())
				.build();
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of(group));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogOranigastionsEinheit));
		when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.SUB_GROUP_ORGANISATIONS_EINHEIT_ID)).thenReturn(
				List.of(pvogChildOranigastionsEinheit));

		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp);
		when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of());
		service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp + 1000);

		var parentSynced = repository.findSyncedByKeycloakId(group.getId()).get();
		var childSynced = repository.findSyncedByKeycloakId(subGroup.getId()).get();

		assertThat(parentSynced)
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						group.getId(),
						group.getName(),
						group.getOrganisationsEinheitId(),
						SyncResult.DELETED,
						pvogOranigastionsEinheit.getId(),
						null,
						Collections.emptyList(),
						syncTimestamp);
		assertThat(childSynced)
				.matches(organisationsEinheit -> Objects.nonNull(organisationsEinheit.getId()))
				.matches(organisationsEinheit -> Objects.isNull(organisationsEinheit.getSettings().getSignatur()))
				.extracting(
						OrganisationsEinheit::getKeycloakId,
						OrganisationsEinheit::getName,
						OrganisationsEinheit::getOrganisationsEinheitId,
						OrganisationsEinheit::getSyncResult,
						OrganisationsEinheit::getZufiId,
						OrganisationsEinheit::getParentId,
						OrganisationsEinheit::getChildren,
						OrganisationsEinheit::getLastSyncTimestamp
				).containsExactly(
						subGroup.getId(),
						subGroup.getName(),
						subGroup.getOrganisationsEinheitId(),
						SyncResult.DELETED,
						pvogChildOranigastionsEinheit.getId(),
						parentSynced.getId(),
						Collections.emptyList(),
						syncTimestamp);
	}

}