/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.organisationseinheit;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.Instant;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.admin.keycloak.AddGroupData;
import de.ozgcloud.admin.keycloak.AddGroupDataTestFactory;
import de.ozgcloud.admin.keycloak.Group;
import de.ozgcloud.admin.keycloak.GroupTestFactory;
import de.ozgcloud.admin.keycloak.KeycloakRemoteService;
import de.ozgcloud.admin.keycloak.ResourceCreationException;

class SyncServiceTest {

	@Spy
	@InjectMocks
	private SyncService service;

	@Mock
	private OrganisationsEinheitRepository repository;

	@Mock
	private KeycloakRemoteService keycloakRemoteService;

	@Mock
	private OrganisationsEinheitRemoteService organisationsEinheitRemoteService;

	@Mock
	private OrganisationsEinheitMapper organisationsEinheitMapper;

	@Nested
	class TestSyncOrganisationsEinheitenFromKeycloak {

		private final long syncTimestamp = Instant.now().toEpochMilli();
		private final Group group = GroupTestFactory.create();

		@BeforeEach
		void setUp() {
			when(keycloakRemoteService.getGroupsWithOrganisationsEinheitId()).thenReturn(Stream.of(group));
			doNothing().when(service).syncGroupsWithSubGroups(any(), any(), anyLong());
		}

		@Test
		void shouldGetGroupsFromKeycloak() {
			callService();

			verify(keycloakRemoteService).getGroupsWithOrganisationsEinheitId();
		}

		@Test
		void shouldSyncGroupsFromKeycloak() {
			callService();

			verify(service).syncGroupsWithSubGroups(group, null, syncTimestamp);
		}

		@Test
		void shouldSetUnsyncedAsDeleted() {
			callService();

			verify(repository).setUnsyncedAsDeleted(syncTimestamp);
		}

		private void callService() {
			service.syncOrganisationsEinheitenFromKeycloak(syncTimestamp);
		}

	}

	@Nested
	class TestSyncGroupsWithSubGroups {

		private final long syncTimestamp = Instant.now().toEpochMilli();
		private final Group group = GroupTestFactory.create();
		private final OrganisationsEinheit syncedOrganisationsEinheitGroup = OrganisationsEinheitTestFactory.create();
		private final OrganisationsEinheit savedOrganisationsEinheitGroup = OrganisationsEinheitTestFactory.createBuilder()
				.id(UUID.randomUUID().toString()).build();

		@BeforeEach
		void setUp() {
			doReturn(syncedOrganisationsEinheitGroup).when(service).syncGroup(group, null, syncTimestamp);
			doReturn(savedOrganisationsEinheitGroup).when(service).saveSyncedOrganisationsEinheit(syncedOrganisationsEinheitGroup);
		}

		@Test
		void shouldSyncGroup() {
			service.syncGroupsWithSubGroups(group, null, syncTimestamp);

			verify(service).syncGroup(group, null, syncTimestamp);
		}

		@Test
		void shouldSaveSyncedOrganisationsEinheit() {
			service.syncGroupsWithSubGroups(group, null, syncTimestamp);

			verify(service).saveSyncedOrganisationsEinheit(syncedOrganisationsEinheitGroup);
		}

		@Test
		void shouldSyncSubGroupsWithSyncedOrganisationsEinheitAsParent() {
			service.syncGroupsWithSubGroups(group, null, syncTimestamp);

			verify(service).syncGroupsWithSubGroups(group.getSubGroups().getFirst(), savedOrganisationsEinheitGroup, syncTimestamp);
		}
	}

	@Nested
	class TestSyncGroup {

		private final long syncTimestamp = Instant.now().toEpochMilli();
		private final Group group = GroupTestFactory.create();
		private final OrganisationsEinheit parent = OrganisationsEinheitTestFactory.createBuilder().id(UUID.randomUUID().toString()).build();
		private final OrganisationsEinheit pvogOrganisationsEinheit = OrganisationsEinheitTestFactory.createBuilder().zufiId(null).settings(null)
				.parentId(null).syncResult(null).build();
		private final String syncedName = LoremIpsum.getInstance().getWords(3);
		private final SyncResult syncedSyncResult = SyncResult.OK;
		private final String syncedZufiId = UUID.randomUUID().toString();

		@Nested
		class ParentGroup {

			@BeforeEach
			void setUp() {
				doReturn(syncedName).when(service).syncName(List.of(pvogOrganisationsEinheit), group);
				doReturn(syncedSyncResult).when(service).evaluateSyncResult(List.of(pvogOrganisationsEinheit), group);
				doReturn(syncedZufiId).when(service).syncZufiId(List.of(pvogOrganisationsEinheit));
				when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(
						List.of(pvogOrganisationsEinheit));
			}

			@Test
			void shouldSyncKeycloakId() {
				var synced = service.syncGroup(group, null, syncTimestamp);

				assertThat(synced.getKeycloakId()).isEqualTo(GroupTestFactory.ID);
			}

			@Test
			void shouldSyncName() {
				var synced = service.syncGroup(group, null, syncTimestamp);

				verify(service).syncName(List.of(pvogOrganisationsEinheit), group);
				assertThat(synced.getName()).isEqualTo(syncedName);
			}

			@Test
			void shouldEvaluateSyncResult() {
				var synced = service.syncGroup(group, null, syncTimestamp);

				verify(service).evaluateSyncResult(List.of(pvogOrganisationsEinheit), group);
				assertThat(synced.getSyncResult()).isEqualTo(syncedSyncResult);
			}

			@Test
			void shouldGetOrganisationsEinheit() {
				service.syncGroup(group, null, syncTimestamp);

				verify(organisationsEinheitRemoteService).getByOrganisationsEinheitId(GroupTestFactory.ORGANISATIONS_EINHEIT_ID);
			}

			@Test
			void shouldSetOrganisationsEinheitId() {
				var synced = service.syncGroup(group, null, syncTimestamp);

				assertThat(synced.getOrganisationsEinheitId()).isEqualTo(group.getOrganisationsEinheitId());
			}

			@Test
			void shouldSyncZufiId() {
				var synced = service.syncGroup(group, null, syncTimestamp);

				verify(service).syncZufiId(List.of(pvogOrganisationsEinheit));
				assertThat(synced.getZufiId()).isEqualTo(syncedZufiId);
			}

			@Test
			void shouldSetLastSyncTimestamp() {
				var synced = service.syncGroup(group, null, syncTimestamp);

				assertThat(synced.getLastSyncTimestamp()).isEqualTo(syncTimestamp);
			}
		}

		@Nested
		class SubGroup {

			@BeforeEach
			void setUp() {
				doReturn(syncedName).when(service).syncName(List.of(pvogOrganisationsEinheit), group.getSubGroups().getFirst());
				doReturn(syncedSyncResult).when(service).evaluateSyncResult(List.of(pvogOrganisationsEinheit), group.getSubGroups().getFirst());
				doReturn(syncedZufiId).when(service).syncZufiId(List.of(pvogOrganisationsEinheit));
				when(organisationsEinheitRemoteService.getByOrganisationsEinheitId(GroupTestFactory.SUB_GROUP_ORGANISATIONS_EINHEIT_ID)).thenReturn(
						List.of(pvogOrganisationsEinheit));
			}

			@Test
			void shouldSyncKeycloakId() {
				var synced = service.syncGroup(group.getSubGroups().getFirst(), parent, syncTimestamp);

				assertThat(synced.getKeycloakId()).isEqualTo(GroupTestFactory.SUB_GROUP_ID);
			}

			@Test
			void shouldSyncName() {
				var synced = service.syncGroup(group.getSubGroups().getFirst(), parent, syncTimestamp);

				verify(service).syncName(List.of(pvogOrganisationsEinheit), group.getSubGroups().getFirst());
				assertThat(synced.getName()).isEqualTo(syncedName);
			}

			@Test
			void shouldEvaluateSyncResult() {
				var synced = service.syncGroup(group.getSubGroups().getFirst(), parent, syncTimestamp);

				verify(service).evaluateSyncResult(List.of(pvogOrganisationsEinheit), group.getSubGroups().getFirst());
				assertThat(synced.getSyncResult()).isEqualTo(syncedSyncResult);
			}

			@Test
			void shouldGetOrganisationsEinheit() {
				var synced = service.syncGroup(group.getSubGroups().getFirst(), parent, syncTimestamp);

				verify(organisationsEinheitRemoteService).getByOrganisationsEinheitId(GroupTestFactory.SUB_GROUP_ORGANISATIONS_EINHEIT_ID);
				assertThat(synced.getOrganisationsEinheitId()).isEqualTo(group.getSubGroups().getFirst().getOrganisationsEinheitId());
			}

			@Test
			void shouldSetParent() {
				var synced = service.syncGroup(group.getSubGroups().getFirst(), parent, syncTimestamp);

				assertThat(synced.getParentId()).contains(parent.getId());
			}

			@Test
			void shouldSyncZufiId() {
				var synced = service.syncGroup(group.getSubGroups().getFirst(), parent, syncTimestamp);

				verify(service).syncZufiId(List.of(pvogOrganisationsEinheit));
				assertThat(synced.getZufiId()).isEqualTo(syncedZufiId);
			}

			@Test
			void shouldSetLastSyncTimestamp() {
				var synced = service.syncGroup(group.getSubGroups().getFirst(), parent, syncTimestamp);

				assertThat(synced.getLastSyncTimestamp()).isEqualTo(syncTimestamp);
			}
		}
	}

	@Nested
	class TestSyncName {

		@Test
		void shouldReturnPvogName() {
			var name = service.syncName(List.of(OrganisationsEinheitTestFactory.create()), GroupTestFactory.create());

			assertThat(name).isEqualTo(OrganisationsEinheitTestFactory.NAME);

		}

		@Test
		void shouldReturnGroupNameIfOrganisatonsEinheitNotFoundInPvog() {
			var name = service.syncName(Collections.emptyList(), GroupTestFactory.create());

			assertThat(name).isEqualTo(GroupTestFactory.NAME);
		}

		@Test
		void shouldReturnGroupNameIfMultipleOrganisatonsEinheitenFoundInPvog() {
			var name = service.syncName(List.of(OrganisationsEinheitTestFactory.create(), OrganisationsEinheitTestFactory.create()),
					GroupTestFactory.create());

			assertThat(name).isEqualTo(GroupTestFactory.NAME);
		}
	}

	@Nested
	class TestEvaluateSyncResult {

		@Test
		void shouldReturnOk() {
			var syncResult = service.evaluateSyncResult(List.of(OrganisationsEinheitTestFactory.createBuilder().name(GroupTestFactory.NAME).build()),
					GroupTestFactory.create());

			assertThat(syncResult).isEqualTo(SyncResult.OK);
		}

		@Test
		void shouldReturnNotFoundInPvog() {
			var syncResult = service.evaluateSyncResult(Collections.emptyList(), GroupTestFactory.create());

			assertThat(syncResult).isEqualTo(SyncResult.NOT_FOUND_IN_PVOG);
		}

		@Test
		void shouldReturnNameMismatch() {
			var syncResult = service.evaluateSyncResult(List.of(OrganisationsEinheitTestFactory.create()), GroupTestFactory.create());

			assertThat(syncResult).isEqualTo(SyncResult.NAME_MISMATCH);
		}

		@Test
		void shouldReturnOrganisationseinheitIdNotUnique() {
			var syncResult = service.evaluateSyncResult(List.of(OrganisationsEinheitTestFactory.create(), OrganisationsEinheitTestFactory.create()),
					GroupTestFactory.create());

			assertThat(syncResult).isEqualTo(SyncResult.ORGANISATIONSEINHEIT_ID_NOT_UNIQUE);
		}

	}

	@Nested
	class TestSyncZufiId {

		@Test
		void shouldReturnZufiId() {
			var zufiId = service.syncZufiId(List.of(OrganisationsEinheitTestFactory.create()));

			assertThat(zufiId).isEqualTo(OrganisationsEinheitTestFactory.ID);
		}

		@Test
		void shouldReturnNullIfOrganisationsEinheitNotInPvog() {
			var zufiId = service.syncZufiId(Collections.emptyList());

			assertThat(zufiId).isNull();
		}

		@Test
		void shouldReturnNullIfOrganisationsEinheitNotUnique() {
			var zufiId = service.syncZufiId(List.of(OrganisationsEinheitTestFactory.create(), OrganisationsEinheitTestFactory.create()));

			assertThat(zufiId).isNull();
		}
	}

	@Nested
	class TestSaveSyncedOrganisationsEinheit {

		private final OrganisationsEinheit syncedOrganisationsEinheit = OrganisationsEinheitTestFactory.createBuilder().id(null).settings(null)
				.build();
		private final OrganisationsEinheit savedOrganisationsEinheit = OrganisationsEinheitTestFactory.createBuilder().settings(null)
				.build();

		@Test
		void shouldFindSyncedOrganisationsEinheitInDatabase() {
			service.saveSyncedOrganisationsEinheit(syncedOrganisationsEinheit);

			verify(repository).findSyncedByKeycloakId(syncedOrganisationsEinheit.getKeycloakId());
		}

		@Test
		void shouldReturnSaved() {
			when(repository.save(syncedOrganisationsEinheit)).thenReturn(savedOrganisationsEinheit);

			var saved = service.saveSyncedOrganisationsEinheit(syncedOrganisationsEinheit);

			assertThat(saved).isEqualTo(savedOrganisationsEinheit);
		}

		@Nested
		class SyncedOrganisationsEinheitDoesNotExist {

			@BeforeEach
			void setUp() {
				when(repository.findSyncedByKeycloakId(
						syncedOrganisationsEinheit.getKeycloakId())).thenReturn(Optional.empty());
			}

			@Test
			void shouldSave() {
				service.saveSyncedOrganisationsEinheit(syncedOrganisationsEinheit);

				verify(repository).save(syncedOrganisationsEinheit);
			}
		}

		@Nested
		class SyncedOrganisationsEinheitExists {

			private final OrganisationsEinheit existingOrganisationsEinheit = OrganisationsEinheitTestFactory.createNewBuilder().build();

			@Captor
			private ArgumentCaptor<OrganisationsEinheit> savedOrganisationsEinheitArgumentCaptor;

			@BeforeEach
			void setUp() {
				when(repository.findSyncedByKeycloakId(
						syncedOrganisationsEinheit.getKeycloakId())).thenReturn(Optional.of(existingOrganisationsEinheit));
			}

			@Test
			void shouldSaveSyncedOrganisationsEinheitWithSettings() {
				service.saveSyncedOrganisationsEinheit(syncedOrganisationsEinheit);

				verify(repository).save(savedOrganisationsEinheitArgumentCaptor.capture());
				assertThat(savedOrganisationsEinheitArgumentCaptor.getValue())
						.extracting(
								OrganisationsEinheit::getId,
								OrganisationsEinheit::getOrganisationsEinheitId,
								OrganisationsEinheit::getSettings,
								OrganisationsEinheit::getName,
								OrganisationsEinheit::getSyncResult,
								OrganisationsEinheit::getParentId,
								OrganisationsEinheit::getZufiId)
						.containsExactly(
								existingOrganisationsEinheit.getId(),
								syncedOrganisationsEinheit.getOrganisationsEinheitId(),
								existingOrganisationsEinheit.getSettings(),
								syncedOrganisationsEinheit.getName(),
								syncedOrganisationsEinheit.getSyncResult(),
								syncedOrganisationsEinheit.getParentId(),
								syncedOrganisationsEinheit.getZufiId()
						);
			}
		}
	}

	@Nested
	class TestSyncAddedOrganisationsEinheiten {

		private final long syncTimestamp = Instant.now().toEpochMilli();
		private final OrganisationsEinheit withoutSyncResult1 = OrganisationsEinheitTestFactory.createBuilder().id("A").build();
		private final OrganisationsEinheit withoutSyncResult2 = OrganisationsEinheitTestFactory.createBuilder().id("B").build();
		private final OrganisationsEinheit[] organisationsEinheiten = new OrganisationsEinheit[] { withoutSyncResult2, withoutSyncResult1 };

		@Captor
		private ArgumentCaptor<Stream<OrganisationsEinheit>> streamArgumentCaptor;

		@BeforeEach
		void setUp() {
			when(repository.findAllWithoutSyncResult()).thenReturn(Stream.of(organisationsEinheiten));
			doNothing().when(service).syncAddedOrganisationsEinheit(any(), anyLong());
		}

		@Test
		void shouldFindAllWithoutSyncResult() {
			callService();

			verify(repository).findAllWithoutSyncResult();
		}

		@Test
		void shouldSynchronizeFirstOrganisationsEinheit() {
			callService();

			verify(service).syncAddedOrganisationsEinheit(withoutSyncResult1, syncTimestamp);
		}

		@Test
		void shouldSynchronizeSecondOrganisationsEinheit() {
			callService();

			verify(service).syncAddedOrganisationsEinheit(withoutSyncResult2, syncTimestamp);
		}

		private void callService() {
			service.syncAddedOrganisationsEinheiten(syncTimestamp);
		}
	}

	@Nested
	class TestSyncAddedOrganisationsEinheit {

		private final long syncTimestamp = Instant.now().toEpochMilli();
		private final String addedGroupId = UUID.randomUUID().toString();
		private final OrganisationsEinheit organisationsEinheit = OrganisationsEinheitTestFactory.create();

		@Test
		void shouldAddAsGroupInKeycloak() {
			doReturn(Optional.empty()).when(service).addAsGroupInKeycloak(organisationsEinheit);

			callService();

			verify(service).addAsGroupInKeycloak(organisationsEinheit);
		}

		@Test
		void shouldUpdateAfterSuccessfulGroupCreation() {
			doReturn(Optional.of(addedGroupId)).when(service).addAsGroupInKeycloak(organisationsEinheit);
			doNothing().when(service).updateAfterSuccessfulGroupCreation(organisationsEinheit, addedGroupId, syncTimestamp);

			callService();

			verify(service).updateAfterSuccessfulGroupCreation(organisationsEinheit, addedGroupId, syncTimestamp);
		}

		@Test
		void shouldNotUpdate() {
			doReturn(Optional.empty()).when(service).addAsGroupInKeycloak(organisationsEinheit);

			callService();

			verify(service, never()).updateAfterSuccessfulGroupCreation(any(), any(), anyLong());
		}

		private void callService() {
			service.syncAddedOrganisationsEinheit(organisationsEinheit, syncTimestamp);
		}
	}

	@Nested
	class TestAddAsGroupInKeycloak {

		@Nested
		class OnParentIdIsNotSet {

			private final OrganisationsEinheit organisationsEinheit = OrganisationsEinheitTestFactory.createBuilder().parentId(null).build();
			private final AddGroupData addGroupData = AddGroupDataTestFactory.create();
			private final String addedGroupId = UUID.randomUUID().toString();

			@BeforeEach
			void init() {
				when(organisationsEinheitMapper.toAddGroupData(organisationsEinheit)).thenReturn(addGroupData);
				doReturn(Optional.of(addedGroupId)).when(service).addGroupInKeycloak(addGroupData);
			}

			@Test
			void shouldCreateAddGroupData() {
				callService();

				verify(organisationsEinheitMapper).toAddGroupData(organisationsEinheit);
			}

			@Test
			void shouldAddGroupInKeycloak() {
				callService();

				verify(service).addGroupInKeycloak(addGroupData);
			}

			@Test
			void shouldReturnAddedGroupId() {
				var returnedGroupId = callService();

				assertThat(returnedGroupId).get().isEqualTo(addedGroupId);
			}

			private Optional<String> callService() {
				return service.addAsGroupInKeycloak(organisationsEinheit);
			}
		}

		@Nested
		class OnParentIdIsSet {

			@Test
			void shouldThrowException() {
				var withParentId = OrganisationsEinheitTestFactory.create();
				assertThatExceptionOfType(OrganisationsEinheitSynchronizationException.class)
						.isThrownBy(() -> service.addAsGroupInKeycloak(withParentId))
						.withMessage("Organisationseinheit " + withParentId.getOrganisationsEinheitId() + " has parent");
			}
		}
	}

	@Nested
	class TestUpdateAfterSuccessfulGroupCreation {

		private final OrganisationsEinheit organisationsEinheit = OrganisationsEinheitTestFactory.createBuilder()
				.keycloakId(null)
				.syncResult(null)
				.lastSyncTimestamp(null)
				.build();
		private final long syncTimestamp = Instant.now().toEpochMilli();
		private final String keycloakId = UUID.randomUUID().toString();

		@Captor
		private ArgumentCaptor<OrganisationsEinheit> organisationsEinheitArgumentCaptor;

		@Test
		void shouldSaveUpdatedOrganisationsEinheit() {
			service.updateAfterSuccessfulGroupCreation(organisationsEinheit, keycloakId, syncTimestamp);

			verify(repository).save(organisationsEinheitArgumentCaptor.capture());
			assertThat(organisationsEinheitArgumentCaptor.getValue())
					.extracting(OrganisationsEinheit::getKeycloakId, OrganisationsEinheit::getSyncResult, OrganisationsEinheit::getLastSyncTimestamp)
					.containsExactly(keycloakId, SyncResult.OK, syncTimestamp);
		}
	}

	@Nested
	class TestAddGroupInKeycloak {

		private static final String FAILURE_MESSAGE = LoremIpsum.getInstance().getWords(5);
		private final Throwable resourceCreationException = new ResourceCreationException(FAILURE_MESSAGE);

		private final String keycloakId = GroupTestFactory.ID;
		private final AddGroupData addGroupData = AddGroupDataTestFactory.create();

		@Test
		void shouldAddGroup() {
			givenAddGroupSuccessful();

			callService();

			verify(keycloakRemoteService).addGroup(addGroupData);
		}

		@Test
		void shouldReturnKeycloakId() {
			givenAddGroupSuccessful();

			var keycloakId = callService();

			assertThat(keycloakId).isPresent().get().isEqualTo(this.keycloakId);
		}

		@Test
		void shouldThrowExceptionInCaseOfResourceCreationException() {
			givenAddGroupFailed();

			assertThatExceptionOfType(OrganisationsEinheitSynchronizationException.class).isThrownBy(this::callService)
					.withMessage("Error creating group %s in Keycloak", addGroupData.toString()).withCause(resourceCreationException);
		}

		private void givenAddGroupSuccessful() {
			when(keycloakRemoteService.addGroup(addGroupData)).thenReturn(keycloakId);
		}

		private void givenAddGroupFailed() {
			when(keycloakRemoteService.addGroup(addGroupData)).thenThrow(resourceCreationException);
		}

		private Optional<String> callService() {
			return service.addGroupInKeycloak(addGroupData);
		}
	}

}