/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.security;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.net.URI;
import java.util.List;

import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ProblemDetail;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.web.ErrorResponse;
import org.springframework.web.servlet.function.ServerResponse;
import org.springframework.web.servlet.function.ServerResponse.Context;

import lombok.SneakyThrows;

class AdminAuthenticationEntryPointTest {

	private static final String REQUEST_URI = "/request-uri";

	@Spy
	@InjectMocks
	private AdminAuthenticationEntryPoint authenticationEntryPoint;

	@Mock
	private List<HttpMessageConverter<?>> converters;

	@Mock
	private HttpServletRequest request;

	@DisplayName("commence")
	@Nested
	class TestCommence {
		@Mock
		private HttpServletResponse response;

		@Mock
		private ServerResponse serverResponse;

		@BeforeEach
		void initMock() {
			when(request.getRequestURI()).thenReturn(REQUEST_URI);
		}

		@Captor
		ArgumentCaptor<Context> contextCaptor;

		@DisplayName("call writeTo")
		@SneakyThrows
		@Test
		void shouldCallWriteTo() {
			var authException = AuthenticationExceptionTestFactory.create();
			try (var mockServerResponse = mockStatic(ServerResponse.class)) {
				mockServerResponse.when(() -> ServerResponse.from(any(ErrorResponse.class))).thenReturn(serverResponse);

				authenticationEntryPoint.commence(request, response, authException);

				verify(serverResponse).writeTo(eq(request), eq(response), contextCaptor.capture());
				assertThat(contextCaptor.getValue().messageConverters()).isEqualTo(converters);
			}
		}
	}

	@DisplayName("error response")
	@Nested
	class TestBuildErrorResponse {

		@Nested
		class TestBody {
			@Test
			void shouldHaveInstance() {
				var problemDetail = getErrorResponseBody();

				assertThat(problemDetail.getInstance()).isEqualTo(URI.create(REQUEST_URI));
			}

			@Test
			void shouldHaveDetail() {
				var problemDetail = getErrorResponseBody();

				assertThat(problemDetail.getDetail()).isEqualTo(AuthenticationExceptionTestFactory.AUTH_MESSAGE);
			}

			private ProblemDetail getErrorResponseBody() {
				return getServerErrorResponse().getBody();
			}

		}

		@Nested
		class TestHeaders {

			@Test
			void shouldHaveContentType() {
				var headers = getHttpHeaders();

				assertThat(headers.getContentType()).isEqualTo(MediaType.APPLICATION_PROBLEM_JSON);
			}

			@Test
			void shouldHaveWwwAuthentication() {
				var headers = getHttpHeaders();

				assertThat(headers.getFirst(HttpHeaders.WWW_AUTHENTICATE)).isEqualTo("Bearer realm=\"Restricted Content\"");
			}

			private HttpHeaders getHttpHeaders() {
				return getServerErrorResponse().getHeaders();
			}
		}

		@Test
		void shouldHaveStatus() {
			var response = getServerErrorResponse();

			assertThat(response.getStatusCode()).isEqualTo(HttpStatus.UNAUTHORIZED);
		}

		private ErrorResponse getServerErrorResponse() {
			return authenticationEntryPoint.buildErrorResponse(
					AuthenticationExceptionTestFactory.create(),
					REQUEST_URI);
		}
	}

}
