/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.setting;

import static org.assertj.core.api.Assertions.*;

import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.EnableAutoConfiguration;
import org.springframework.boot.test.autoconfigure.data.mongo.DataMongoTest;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.ContextConfiguration;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.admin.organisationseinheit.OrganisationsEinheit;
import de.ozgcloud.admin.organisationseinheit.OrganisationsEinheitSettings;
import de.ozgcloud.admin.organisationseinheit.OrganisationsEinheitSettingsTestFactory;
import de.ozgcloud.admin.organisationseinheit.OrganisationsEinheitTestFactory;
import de.ozgcloud.common.test.DbInitializer;

@DataMongoTest
@ContextConfiguration(initializers = { DbInitializer.class }, classes = { OrganisationsEinheitSettingsRepository.class, MongoOperations.class })
@ActiveProfiles({ "itcase", "with_db" })
@EnableAutoConfiguration
class OrganisationsEinheitSettingsRepositoryITCase {

	@Autowired
	private OrganisationsEinheitSettingsRepository repository;
	@Autowired
	private MongoOperations operations;

	@BeforeEach
	void clearDatabase() {
		operations.dropCollection(OrganisationsEinheit.class);
	}

	@Nested
	class TestGetSettings {

		private final OrganisationsEinheit withSignatur1 = OrganisationsEinheitTestFactory.createBuilder()
				.id(UUID.randomUUID().toString())
				.organisationsEinheitId(UUID.randomUUID().toString())
				.settings(withRandomSignatur())
				.build();
		private final OrganisationsEinheit withSignatur2 = OrganisationsEinheitTestFactory.createBuilder()
				.id(UUID.randomUUID().toString())
				.organisationsEinheitId(UUID.randomUUID().toString())
				.settings(withRandomSignatur())
				.build();
		private final OrganisationsEinheit withNullSettings = OrganisationsEinheitTestFactory.createBuilder()
				.id(UUID.randomUUID().toString())
				.organisationsEinheitId(UUID.randomUUID().toString())
				.settings(null)
				.build();
		private final OrganisationsEinheit withEmptySettings = OrganisationsEinheitTestFactory.createBuilder()
				.id(UUID.randomUUID().toString())
				.organisationsEinheitId(UUID.randomUUID().toString())
				.settings(OrganisationsEinheitSettings.builder().build())
				.build();

		@BeforeEach
		void setUp() {
			operations.save(withSignatur1);
			operations.save(withSignatur2);
			operations.save(withNullSettings);
			operations.save(withEmptySettings);
		}

		@Test
		void shouldFind() {
			var found = repository.getSettings();

			assertThat(found).extracting(OrganisationsEinheitWithSettings::organisationsEinheitId, item -> item.settings().getSignatur())
					.containsExactlyInAnyOrder(
							tuple(withSignatur1.getOrganisationsEinheitId(), withSignatur1.getSettings().getSignatur()),
							tuple(withSignatur2.getOrganisationsEinheitId(), withSignatur2.getSettings().getSignatur()));
		}

		private static OrganisationsEinheitSettings withRandomSignatur() {
			return withSignatur(LoremIpsum.getInstance().getWords(5));
		}

		private static OrganisationsEinheitSettings withSignatur(String signatur) {
			return OrganisationsEinheitSettingsTestFactory.createBuilder().signatur(signatur).build();
		}
	}
}
